import pytest

import semantic_scholar.api.authors as authors_api


pytestmark = pytest.mark.asyncio


def assert_single_call(
    mock_make_request,
    *,
    endpoint,
    params,
    api_key_override=None,
    method="GET",
    json=None,
    base_url=None,
):
    assert mock_make_request.calls == [
        {
            "endpoint": endpoint,
            "params": params,
            "api_key_override": api_key_override,
            "method": method,
            "json": json,
            "base_url": base_url,
        }
    ]


def assert_validation_error(result, message, details=None):
    assert result == {
        "error": {
            "type": "validation",
            "message": message,
            "details": details or {},
        }
    }


async def test_author_search_happy_path(mock_make_request):
    payload = {"total": 1, "offset": 10, "data": [{"authorId": "1741101"}]}
    mock_make_request.install(authors_api).queue_responses(payload)

    result = await authors_api.author_search.fn(
        None,
        query="Andrew Ng",
        fields=["name", "paperCount"],
        offset=10,
        limit=25,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/search",
        params={
            "query": "Andrew Ng",
            "offset": 10,
            "limit": 25,
            "fields": "name,paperCount",
        },
    )


async def test_author_search_empty_query_validation(mock_make_request):
    result = await authors_api.author_search.fn(None, query="   ")

    assert_validation_error(result, "Query string cannot be empty")
    assert mock_make_request.calls == []


async def test_author_search_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(authors_api).queue_responses(error)

    result = await authors_api.author_search.fn(
        None,
        query="Andrew Ng",
        fields=["name", "paperCount"],
        offset=10,
        limit=25,
    )

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/author/search",
        params={
            "query": "Andrew Ng",
            "offset": 10,
            "limit": 25,
            "fields": "name,paperCount",
        },
    )


async def test_author_details_happy_path(mock_make_request):
    payload = {"authorId": "1741101", "name": "Andrew Ng"}
    mock_make_request.install(authors_api).queue_responses(payload)

    result = await authors_api.author_details.fn(None, author_id="1741101", fields=["name", "hIndex"])

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/1741101",
        params={"fields": "name,hIndex"},
    )


async def test_author_details_empty_author_id_validation(mock_make_request):
    result = await authors_api.author_details.fn(None, author_id="   ")

    assert_validation_error(result, "Author ID cannot be empty")
    assert mock_make_request.calls == []


async def test_author_details_404_maps_to_author_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(authors_api).queue_responses(mock_error_response(status_code=404))

    result = await authors_api.author_details.fn(None, author_id="1741101", fields=["name", "hIndex"])

    assert_validation_error(result, "Author not found", {"author_id": "1741101"})
    assert_single_call(
        mock_make_request,
        endpoint="/author/1741101",
        params={"fields": "name,hIndex"},
    )


async def test_author_papers_happy_path(mock_make_request):
    payload = {"offset": 0, "next": 10, "data": [{"paperId": "p1"}]}
    mock_make_request.install(authors_api).queue_responses(payload)

    result = await authors_api.author_papers.fn(
        None,
        author_id="1741101",
        fields=["title", "year"],
        offset=0,
        limit=10,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/1741101/papers",
        params={"offset": 0, "limit": 10, "fields": "title,year"},
    )


async def test_author_papers_limit_validation(mock_make_request):
    result = await authors_api.author_papers.fn(None, author_id="1741101", limit=1001)

    assert_validation_error(result, "Limit cannot exceed 1000", {"max_limit": 1000})
    assert mock_make_request.calls == []


async def test_author_papers_404_maps_to_author_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(authors_api).queue_responses(mock_error_response(status_code=404))

    result = await authors_api.author_papers.fn(
        None,
        author_id="1741101",
        fields=["title", "year"],
        offset=0,
        limit=10,
    )

    assert_validation_error(result, "Author not found", {"author_id": "1741101"})
    assert_single_call(
        mock_make_request,
        endpoint="/author/1741101/papers",
        params={"offset": 0, "limit": 10, "fields": "title,year"},
    )


async def test_author_batch_details_happy_path(mock_make_request):
    payload = [{"authorId": "1741101"}, {"authorId": "2061296"}]
    mock_make_request.install(authors_api).queue_responses(payload)

    result = await authors_api.author_batch_details.fn(
        None,
        author_ids=["1741101", "2061296"],
        fields="name,paperCount",
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/batch",
        params={"fields": "name,paperCount"},
        method="POST",
        json={"ids": ["1741101", "2061296"]},
    )


async def test_author_batch_details_empty_ids_validation(mock_make_request):
    result = await authors_api.author_batch_details.fn(None, author_ids=[], fields="name,paperCount")

    assert_validation_error(result, "Author IDs list cannot be empty")
    assert mock_make_request.calls == []


async def test_author_batch_details_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(authors_api).queue_responses(error)

    result = await authors_api.author_batch_details.fn(
        None,
        author_ids=["1741101", "2061296"],
        fields="name,paperCount",
    )

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/author/batch",
        params={"fields": "name,paperCount"},
        method="POST",
        json={"ids": ["1741101", "2061296"]},
    )



"""
Centralized logging configuration for the Semantic Scholar server.
"""

import logging
import os

DEBUG_REQUESTS = os.getenv("SEMANTIC_SCHOLAR_DEBUG", "").strip().lower() in ("1", "true", "yes", "on")

logging.basicConfig(level=logging.DEBUG if DEBUG_REQUESTS else logging.INFO)
logger = logging.getLogger("semantic_scholar")



import pytest

import semantic_scholar.bridge as bridge
from semantic_scholar.config import AuthorDetailFields, Config, PaperFields


pytestmark = pytest.mark.asyncio


def assert_single_call(
    mock_make_request,
    *,
    endpoint,
    params,
    api_key_override=None,
    method="GET",
    json=None,
    base_url=None,
):
    assert mock_make_request.calls == [
        {
            "endpoint": endpoint,
            "params": params,
            "api_key_override": api_key_override,
            "method": method,
            "json": json,
            "base_url": base_url,
        }
    ]


async def test_bridge_paper_search_contract(mock_make_request, bridge_client, auth_headers):
    payload = {"total": 1, "data": [{"paperId": "p1"}]}
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.get("/v1/paper/search", params={"q": "attention"}, headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search",
        params={
            "query": "attention",
            "offset": 0,
            "limit": 10,
            "fields": "title,abstract,year,citationCount,authors,url",
        },
        api_key_override="bridge-token",
    )


async def test_bridge_paper_details_contract(mock_make_request, bridge_client, auth_headers):
    payload = {"paperId": "paper-123", "title": "Attention"}
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.get("/v1/paper/paper-123", headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123",
        params={"fields": ",".join(Config.DEFAULT_FIELDS)},
        api_key_override="bridge-token",
    )


async def test_bridge_paper_batch_contract(mock_make_request, bridge_client, auth_headers):
    payload = [{"paperId": "p1"}, {"paperId": "p2"}]
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.post("/v1/paper/batch", json={"ids": ["p1", "p2"]}, headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/batch",
        params={"fields": ",".join(Config.DEFAULT_FIELDS)},
        api_key_override="bridge-token",
        method="POST",
        json={"ids": ["p1", "p2"]},
    )


async def test_bridge_author_search_contract(mock_make_request, bridge_client, auth_headers):
    payload = {"total": 1, "data": [{"authorId": "1741101"}]}
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.get("/v1/author/search", params={"q": "Andrew Ng"}, headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/search",
        params={
            "query": "Andrew Ng",
            "offset": 0,
            "limit": 10,
            "fields": ",".join(AuthorDetailFields.BASIC),
        },
        api_key_override="bridge-token",
    )


async def test_bridge_author_details_contract(mock_make_request, bridge_client, auth_headers):
    payload = {"authorId": "1741101", "name": "Andrew Ng"}
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.get("/v1/author/1741101", headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/1741101",
        params={"fields": ",".join(AuthorDetailFields.BASIC)},
        api_key_override="bridge-token",
    )


async def test_bridge_author_batch_contract(mock_make_request, bridge_client, auth_headers):
    payload = [{"authorId": "1741101"}, {"authorId": "2061296"}]
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.post(
        "/v1/author/batch",
        json={"ids": ["1741101", "2061296"]},
        headers=auth_headers,
    )

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/author/batch",
        params={"fields": ",".join(AuthorDetailFields.BASIC)},
        api_key_override="bridge-token",
        method="POST",
        json={"ids": ["1741101", "2061296"]},
    )


async def test_bridge_recommendations_contract(mock_make_request, bridge_client, auth_headers):
    payload = {"recommendedPapers": [{"paperId": "p9"}]}
    mock_make_request.install(bridge).queue_responses(payload)

    response = await bridge_client.get("/v1/recommendations", params={"paper_id": "paper-123"}, headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == payload
    assert_single_call(
        mock_make_request,
        endpoint="/papers/forpaper/paper-123",
        params={"fields": ",".join(PaperFields.DEFAULT)},
        api_key_override="bridge-token",
        base_url=Config.RECOMMENDATIONS_BASE_URL,
    )


async def test_bridge_recommendations_requires_paper_id(mock_make_request, bridge_client, auth_headers):
    response = await bridge_client.get("/v1/recommendations", headers=auth_headers)

    assert response.status_code == 400
    assert response.json() == {"detail": "paper_id is required"}
    assert mock_make_request.calls == []


async def test_bridge_error_payload_passthrough(mock_make_request, mock_error_response, bridge_client, auth_headers):
    error = mock_error_response(status_code=404)
    mock_make_request.install(bridge).queue_responses(error)

    response = await bridge_client.get("/v1/paper/search", params={"q": "attention"}, headers=auth_headers)

    assert response.status_code == 200
    assert response.json() == error
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search",
        params={
            "query": "attention",
            "offset": 0,
            "limit": 10,
            "fields": "title,abstract,year,citationCount,authors,url",
        },
        api_key_override="bridge-token",
    )



"""Shared core transport primitives for Semantic Scholar clients."""

from .client import S2Client, SupportsRequestJson, get_default_client, make_compat_client
from .exceptions import (
    S2ApiError,
    S2Error,
    S2NotFoundError,
    S2RateLimitError,
    S2TimeoutError,
    S2ValidationError,
)
from .requests import (
    AuthorBatchDetailsRequest,
    AuthorDetailsRequest,
    AuthorPapersRequest,
    AuthorSearchRequest,
    PaperAutocompleteRequest,
    PaperAuthorsRequest,
    PaperBatchDetailsRequest,
    PaperBulkSearchRequest,
    PaperCitationsRequest,
    PaperDetailsRequest,
    PaperRecommendationsMultiRequest,
    PaperRecommendationsSingleRequest,
    PaperReferencesRequest,
    PaperRelevanceSearchRequest,
    PaperTitleSearchRequest,
    RequestModel,
    SnippetSearchRequest,
)
from .transport import (
    MakeRequestCompatTransport,
    RateLimiter,
    S2Transport,
    cleanup_client,
    default_transport,
    error_dict_to_exception,
    get_api_key,
    initialize_client,
    rate_limiter,
)

__all__ = [
    "AuthorBatchDetailsRequest",
    "AuthorDetailsRequest",
    "AuthorPapersRequest",
    "AuthorSearchRequest",
    "MakeRequestCompatTransport",
    "PaperAutocompleteRequest",
    "PaperAuthorsRequest",
    "PaperBatchDetailsRequest",
    "PaperBulkSearchRequest",
    "PaperCitationsRequest",
    "PaperDetailsRequest",
    "PaperRecommendationsMultiRequest",
    "PaperRecommendationsSingleRequest",
    "PaperReferencesRequest",
    "PaperRelevanceSearchRequest",
    "PaperTitleSearchRequest",
    "RateLimiter",
    "RequestModel",
    "S2ApiError",
    "S2Client",
    "S2Error",
    "S2NotFoundError",
    "S2RateLimitError",
    "S2TimeoutError",
    "S2Transport",
    "S2ValidationError",
    "SnippetSearchRequest",
    "SupportsRequestJson",
    "cleanup_client",
    "default_transport",
    "error_dict_to_exception",
    "get_default_client",
    "get_api_key",
    "initialize_client",
    "make_compat_client",
    "rate_limiter",
]



"""Test utilities and core functionality without MCP dependencies"""

import httpx
import logging
import os
from typing import Dict, Optional
import asyncio
from enum import Enum
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

# Basic setup
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class ErrorType(Enum):
    RATE_LIMIT = "rate_limit"
    API_ERROR = "api_error"
    VALIDATION = "validation"
    TIMEOUT = "timeout"

class Config:
    API_VERSION = "v1"
    GRAPH_BASE_URL = f"https://api.semanticscholar.org/graph/{API_VERSION}"
    RECOMMENDATIONS_BASE_URL = "https://api.semanticscholar.org/recommendations/v1"
    TIMEOUT = 30  # seconds

def create_error_response(
    error_type: ErrorType,
    message: str,
    details: Optional[Dict] = None
) -> Dict:
    return {
        "error": {
            "type": error_type.value,
            "message": message,
            "details": details or {}
        }
    }

def get_api_key() -> Optional[str]:
    """Get the Semantic Scholar API key from environment variables."""
    api_key = os.getenv("SEMANTIC_SCHOLAR_API_KEY")
    logger.info(f"API Key found: {'Yes' if api_key else 'No'}")
    return api_key

async def make_request(endpoint: str, params: Dict = None, method: str = "GET", json: Dict = None) -> Dict:
    """Make a request to the Semantic Scholar API."""
    try:
        api_key = get_api_key()
        headers = {"x-api-key": api_key} if api_key else {}
        params = params or {}
        
        # Choose base URL based on endpoint
        is_recommendations = endpoint.startswith("recommendations") or endpoint.startswith("papers/forpaper")
        base_url = Config.RECOMMENDATIONS_BASE_URL if is_recommendations else Config.GRAPH_BASE_URL
        
        # Clean up endpoint
        if endpoint.startswith("/"):
            endpoint = endpoint[1:]
        if is_recommendations and endpoint.startswith("recommendations/"):
            endpoint = endpoint[15:]  # Remove "recommendations/" prefix
            
        url = f"{base_url}/{endpoint}"
        logger.info(f"Making {method} request to {url}")
        logger.info(f"Headers: {headers}")
        logger.info(f"Params: {params}")
        if json:
            logger.info(f"JSON body: {json}")

        async with httpx.AsyncClient(timeout=Config.TIMEOUT, follow_redirects=True) as client:
            if method == "GET":
                response = await client.get(url, params=params, headers=headers)
            else:  # POST
                response = await client.post(url, params=params, json=json, headers=headers)
            
            logger.info(f"Response status: {response.status_code}")
            logger.info(f"Response body: {response.text}")
            
            response.raise_for_status()
            return response.json()

    except httpx.HTTPStatusError as e:
        if e.response.status_code == 429:
            return create_error_response(
                ErrorType.RATE_LIMIT,
                "Rate limit exceeded",
                {"retry_after": e.response.headers.get("retry-after")}
            )
        return create_error_response(
            ErrorType.API_ERROR,
            f"HTTP error: {e.response.status_code}",
            {"response": e.response.text}
        )
    except httpx.TimeoutException:
        return create_error_response(
            ErrorType.TIMEOUT,
            f"Request timed out after {Config.TIMEOUT} seconds"
        )
    except Exception as e:
        return create_error_response(
            ErrorType.API_ERROR,
            str(e)
        ) 


from __future__ import annotations

from typing import Any, Optional, Protocol

from .requests import (
    AuthorBatchDetailsRequest,
    AuthorDetailsRequest,
    AuthorPapersRequest,
    AuthorSearchRequest,
    PaperAutocompleteRequest,
    PaperAuthorsRequest,
    PaperBatchDetailsRequest,
    PaperBulkSearchRequest,
    PaperCitationsRequest,
    PaperDetailsRequest,
    PaperRecommendationsMultiRequest,
    PaperRecommendationsSingleRequest,
    PaperReferencesRequest,
    PaperRelevanceSearchRequest,
    PaperTitleSearchRequest,
    RequestModel,
    SnippetSearchRequest,
)
from .transport import MakeRequestCompatTransport, default_transport


class SupportsRequestJson(Protocol):
    async def request_json(
        self,
        endpoint: str,
        *,
        params: Optional[dict[str, Any]] = None,
        api_key_override: Optional[str] = None,
        method: str = "GET",
        json: Any = None,
        base_url: Optional[str] = None,
    ) -> Any: ...


class S2Client:
    def __init__(self, transport: SupportsRequestJson) -> None:
        self._transport = transport

    async def _request(
        self,
        request: RequestModel,
        *,
        api_key_override: Optional[str] = None,
    ) -> Any:
        return await self._transport.request_json(
            request.endpoint,
            params=request.to_params(),
            api_key_override=api_key_override,
            method=request.method,
            json=request.to_json(),
            base_url=request.base_url,
        )

    async def search_papers(
        self,
        request: PaperRelevanceSearchRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def bulk_search_papers(
        self,
        request: PaperBulkSearchRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def match_paper_title(
        self,
        request: PaperTitleSearchRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_paper(
        self,
        request: PaperDetailsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def batch_papers(
        self,
        request: PaperBatchDetailsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> list[dict[str, Any] | None]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_paper_authors(
        self,
        request: PaperAuthorsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_paper_citations(
        self,
        request: PaperCitationsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_paper_references(
        self,
        request: PaperReferencesRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def autocomplete_papers(
        self,
        request: PaperAutocompleteRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def search_snippets(
        self,
        request: SnippetSearchRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def search_authors(
        self,
        request: AuthorSearchRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_author(
        self,
        request: AuthorDetailsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def get_author_papers(
        self,
        request: AuthorPapersRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def batch_authors(
        self,
        request: AuthorBatchDetailsRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> list[dict[str, Any] | None]:
        return await self._request(request, api_key_override=api_key_override)

    async def recommend_for_paper(
        self,
        request: PaperRecommendationsSingleRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)

    async def recommend_from_papers(
        self,
        request: PaperRecommendationsMultiRequest,
        *,
        api_key_override: Optional[str] = None,
    ) -> dict[str, Any]:
        return await self._request(request, api_key_override=api_key_override)


def get_default_client() -> S2Client:
    return S2Client(default_transport)


def make_compat_client(make_request_callable) -> S2Client:
    return S2Client(MakeRequestCompatTransport(make_request_callable))



"""
Main server module for the Semantic Scholar API Server.
"""

import asyncio
import os
import signal
import uvicorn

# Import mcp from centralized location
from .mcp import mcp
from .utils.http import initialize_client, cleanup_client
from .utils.logger import logger

# Configure logging

# Event to keep process alive when FastMCP detaches
stop_event = None
# ASGI HTTP server instance and task (for the bridge)
http_server = None
http_server_task = None

# Import API modules to register tools
# Note: This must come AFTER mcp is initialized
from .api import papers, authors, recommendations

async def handle_exception(loop, context):
    """Global exception handler for the event loop."""
    msg = context.get("exception", context["message"])
    logger.error(f"Caught exception: {msg}")
    asyncio.create_task(shutdown())

async def shutdown():
    """Gracefully shut down the server."""
    logger.info("Initiating graceful shutdown...")
    
    # Cancel all tasks
    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    for task in tasks:
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
    
    # Cleanup resources
    await cleanup_client()
    try:
        cleanup_fn = getattr(mcp, "cleanup", None)
        if cleanup_fn:
            if asyncio.iscoroutinefunction(cleanup_fn):
                await cleanup_fn()
            else:
                cleanup_fn()
        else:
            # Try common alternative names on FastMCP implementations
            for name in ("shutdown", "stop", "close"):
                fn = getattr(mcp, name, None)
                if fn:
                    if asyncio.iscoroutinefunction(fn):
                        await fn()
                    else:
                        fn()
                    break
    except Exception as e:
        logger.error(f"Error during mcp cleanup: {e}")
    # Signal run_server to stop waiting
    try:
        global stop_event
        if stop_event is not None and not stop_event.is_set():
            stop_event.set()
    except Exception:
        pass

    # Stop the HTTP bridge if running
    try:
        global http_server, http_server_task
        if http_server is not None:
            http_server.should_exit = True
        if http_server_task is not None and not http_server_task.done():
            http_server_task.cancel()
            try:
                await http_server_task
            except asyncio.CancelledError:
                pass
    except Exception as e:
        logger.error(f"Error stopping HTTP bridge: {e}")
    
    logger.info(f"Cancelled {len(tasks)} tasks")
    logger.info("Shutdown complete")

def init_signal_handlers(loop):
    """Initialize signal handlers for graceful shutdown."""
    for sig in (signal.SIGTERM, signal.SIGINT):
        loop.add_signal_handler(sig, lambda: asyncio.create_task(shutdown()))
    logger.info("Signal handlers initialized")

async def run_server():
    """Run the server with proper async context management."""
    try:
        # Initialize HTTP client
        await initialize_client()

        # Start the server
        logger.info("Starting Semantic Scholar Server")
        task = asyncio.create_task(mcp.run_async())

        # Start the HTTP bridge (ASGI) in the same process so the service
        # exposes REST endpoints on a local port. The `bridge.app` is a thin
        # FastAPI application that reuses the package HTTP utilities.
        enable_bridge = os.getenv("SEMANTIC_SCHOLAR_ENABLE_HTTP_BRIDGE", "1").strip().lower() in (
            "1",
            "true",
            "yes",
            "on",
        )
        if enable_bridge:
            bridge_host = os.getenv("SEMANTIC_SCHOLAR_HTTP_BRIDGE_HOST", "0.0.0.0").strip()
            bridge_port = int(os.getenv("SEMANTIC_SCHOLAR_HTTP_BRIDGE_PORT", "8000"))
            from .bridge import app as bridge_app
            config = uvicorn.Config(
                app=bridge_app,
                host=bridge_host,
                port=bridge_port,
                log_level="info",
                log_config=None,
                ws="none"  # Disable WebSocket support to avoid deprecation warnings
            )
            server = uvicorn.Server(config=config)
            global http_server, http_server_task
            http_server = server
            http_server_task = asyncio.create_task(server.serve())
            logger.info("HTTP bridge enabled on %s:%s", bridge_host, bridge_port)
        else:
            logger.info("HTTP bridge disabled (SEMANTIC_SCHOLAR_ENABLE_HTTP_BRIDGE=0)")

        # Create a stop event to keep the main coroutine alive if FastMCP detaches
        global stop_event
        if stop_event is None:
            stop_event = asyncio.Event()

        # Wait until shutdown() sets the event
        await stop_event.wait()
        # Ensure server task is cancelled/awaited on shutdown
        if not task.done():
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass
    except Exception as e:
        logger.error(f"Server error: {e}")
        raise
    finally:
        await shutdown()

def main():
    """Main entry point for the server."""
    try:
        # Set up event loop with exception handler
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_exception_handler(handle_exception)
        
        # Initialize signal handlers
        init_signal_handlers(loop)
        
        # Run the server
        loop.run_until_complete(run_server())
    except KeyboardInterrupt:
        logger.info("Received keyboard interrupt, shutting down...")
    except Exception as e:
        logger.error(f"Fatal error: {str(e)}")
    finally:
        try:
            loop.run_until_complete(asyncio.sleep(0))  # Let pending tasks complete
            loop.close()
        except Exception as e:
            logger.error(f"Error during final cleanup: {str(e)}")
        logger.info("Server stopped")

if __name__ == "__main__":
    main() 



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Optional


@dataclass(eq=False)
class S2Error(Exception):
    message: str
    details: dict[str, Any] = field(default_factory=dict)

    def __str__(self) -> str:
        return self.message


@dataclass(eq=False)
class S2ApiError(S2Error):
    status_code: Optional[int] = None
    endpoint: Optional[str] = None
    method: Optional[str] = None
    params: Optional[dict[str, Any]] = None
    json_body: Any = None
    base_url: Optional[str] = None
    response_text: Optional[str] = None


@dataclass(eq=False)
class S2NotFoundError(S2ApiError):
    resource_type: Optional[str] = None
    resource_id: Optional[str] = None


@dataclass(eq=False)
class S2RateLimitError(S2ApiError):
    retry_after: Optional[str] = None
    authenticated: bool = False


@dataclass(eq=False)
class S2TimeoutError(S2Error):
    endpoint: Optional[str] = None
    method: Optional[str] = None
    timeout_seconds: int = 30


@dataclass(eq=False)
class S2ValidationError(S2Error):
    field: Optional[str] = None



"""
Recommendation-related API endpoints for the Semantic Scholar API.
"""

from typing import Dict, List, Optional

from fastmcp import Context

from ..config import ErrorType
from ..core.client import S2Client, make_compat_client
from ..core.exceptions import S2ApiError, S2Error, S2ValidationError
from ..core.requests import PaperRecommendationsMultiRequest, PaperRecommendationsSingleRequest
from ..mcp import mcp
from ..utils.errors import create_error_response, s2_exception_to_error_response
from ..utils.http import make_request
from ..utils.logger import logger


def _client() -> S2Client:
    return make_compat_client(make_request)


@mcp.tool()
async def get_paper_recommendations_single(
    context: Context,
    paper_id: str,
    fields: Optional[str] = None,
    limit: int = 100,
    from_pool: str = "recent",
) -> Dict:
    try:
        request = PaperRecommendationsSingleRequest(
            paper_id=paper_id,
            fields=fields,
            limit=limit,
            from_pool=from_pool,
        )
        return await _client().recommend_for_paper(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if exc.status_code == 404:
            return create_error_response(
                ErrorType.VALIDATION,
                "Paper not found",
                {"paper_id": paper_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)
    except Exception as exc:
        logger.error("Unexpected error in recommendations: %s", str(exc))
        return create_error_response(
            ErrorType.API_ERROR,
            "Failed to get recommendations",
            {"error": str(exc)},
        )


@mcp.tool()
async def get_paper_recommendations_multi(
    context: Context,
    positive_paper_ids: List[str],
    negative_paper_ids: Optional[List[str]] = None,
    fields: Optional[str] = None,
    limit: int = 100,
) -> Dict:
    try:
        request = PaperRecommendationsMultiRequest(
            positive_paper_ids=positive_paper_ids,
            negative_paper_ids=negative_paper_ids,
            fields=fields,
            limit=limit,
        )
        return await _client().recommend_from_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if exc.status_code == 404:
            return create_error_response(
                ErrorType.VALIDATION,
                "One or more input papers not found",
                {
                    "positive_ids": positive_paper_ids,
                    "negative_ids": negative_paper_ids,
                },
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)
    except Exception as exc:
        logger.error("Unexpected error in recommendations: %s", str(exc))
        return create_error_response(
            ErrorType.API_ERROR,
            "Failed to get recommendations",
            {"error": str(exc)},
        )



from __future__ import annotations

import asyncio
import os
import time
from collections import deque
from typing import Any, Awaitable, Callable, Deque, Dict, Optional, Tuple

import httpx

from ..config import Config, ErrorType, RateLimitConfig
from ..utils.logger import logger
from .exceptions import (
    S2ApiError,
    S2Error,
    S2NotFoundError,
    S2RateLimitError,
    S2TimeoutError,
    S2ValidationError,
)

# Global HTTP client for connection pooling
http_client: Optional[httpx.AsyncClient] = None


class RateLimiter:
    """
    Rate limiter for API requests to prevent exceeding API limits.
    """

    def __init__(
        self,
        *,
        clock: Optional[Callable[[], float]] = None,
        sleeper: Optional[Callable[[float], Awaitable[Any]]] = None,
    ):
        self._clock = clock or time.monotonic
        self._sleep = sleeper or asyncio.sleep
        self._events: Dict[str, Deque[float]] = {}
        self._locks: Dict[str, asyncio.Lock] = {}

    def _bucket_key(self, endpoint: str, base_url: Optional[str] = None) -> str:
        """
        Map a concrete request path to a stable rate-limiting bucket.

        Many endpoints include IDs (e.g. /paper/{id}) which would otherwise
        defeat throttling if used as-is.
        """
        if base_url and "recommendations" in base_url:
            return "/recommendations"
        if "recommendations" in endpoint:
            return "/recommendations"
        if "/author/search" in endpoint:
            return "/author/search"
        if "/paper/search" in endpoint:
            return "/paper/search"
        if "/paper/batch" in endpoint:
            return "/paper/batch"
        if "/author/batch" in endpoint:
            return "/author/batch"
        return "/default"

    def _get_rate_limit(self, endpoint: str, *, authenticated: bool) -> Tuple[int, int]:
        """Get the appropriate rate limit for an endpoint."""
        if not authenticated:
            return RateLimitConfig.UNAUTHENTICATED_LIMIT
        if any(restricted in endpoint for restricted in RateLimitConfig.RESTRICTED_ENDPOINTS):
            if "batch" in endpoint:
                return RateLimitConfig.BATCH_LIMIT
            if "search" in endpoint:
                return RateLimitConfig.SEARCH_LIMIT
            if "recommendations" in endpoint:
                return RateLimitConfig.RECOMMENDATIONS_LIMIT
            return RateLimitConfig.SEARCH_LIMIT
        return RateLimitConfig.DEFAULT_LIMIT

    async def acquire(
        self,
        endpoint: str,
        *,
        authenticated: bool = True,
        base_url: Optional[str] = None,
    ) -> None:
        """
        Acquire permission to make a request, waiting if necessary to respect rate limits.

        Args:
            endpoint: The API endpoint being accessed.
        """
        bucket = self._bucket_key(endpoint, base_url)
        if bucket not in self._locks:
            self._locks[bucket] = asyncio.Lock()
            self._events[bucket] = deque()

        async with self._locks[bucket]:
            limit_endpoint = bucket if bucket != "/default" else endpoint
            requests, seconds = self._get_rate_limit(limit_endpoint, authenticated=authenticated)
            if requests <= 0 or seconds <= 0:
                return

            events = self._events[bucket]
            while True:
                now = self._clock()
                cutoff = now - float(seconds)
                while events and events[0] <= cutoff:
                    events.popleft()

                if len(events) < int(requests):
                    events.append(now)
                    return

                delay = (events[0] + float(seconds)) - now
                if delay > 0:
                    await self._sleep(delay)


rate_limiter = RateLimiter()


def get_api_key() -> Optional[str]:
    """
    Get the Semantic Scholar API key from environment variables.
    Returns None if no API key is set, enabling unauthenticated access.
    """
    api_key = os.getenv("SEMANTIC_SCHOLAR_API_KEY")
    if api_key:
        norm = api_key.strip().lower()
        if norm in ("", "none", "null", "false"):
            logger.warning(
                "SEMANTIC_SCHOLAR_API_KEY is set to a placeholder value; treating as not set."
            )
            return None
        return api_key

    logger.warning("No SEMANTIC_SCHOLAR_API_KEY set. Using unauthenticated access with lower rate limits.")
    return None


async def initialize_client() -> httpx.AsyncClient:
    """Initialize the global HTTP client."""
    global http_client
    if http_client is None:
        http_client = httpx.AsyncClient(
            timeout=Config.TIMEOUT,
            limits=httpx.Limits(max_keepalive_connections=10),
        )
    return http_client


async def cleanup_client() -> None:
    """Clean up the global HTTP client."""
    global http_client
    if http_client is not None:
        await http_client.aclose()
        http_client = None


def _redact_headers(headers: Dict[str, str]) -> Dict[str, str]:
    redacted = dict(headers or {})
    for key in ("x-api-key", "authorization", "proxy-authorization"):
        if key in redacted and redacted[key]:
            redacted[key] = "***"
    return redacted


def _normalize_key(key: Optional[str]) -> Optional[str]:
    if not key:
        return None
    normalized = str(key).strip()
    if normalized.lower() in ("", "none", "null", "false"):
        return None
    return normalized


def error_dict_to_exception(
    result: dict[str, Any],
    *,
    endpoint: Optional[str] = None,
    method: Optional[str] = None,
    params: Optional[dict[str, Any]] = None,
    json_body: Any = None,
    base_url: Optional[str] = None,
) -> S2Error:
    error = result.get("error", {})
    error_type = error.get("type")
    message = error.get("message", "Unknown error")
    details = error.get("details")
    if not isinstance(details, dict):
        details = {}

    if error_type == ErrorType.VALIDATION.value:
        return S2ValidationError(message=message, details=details)

    if error_type == ErrorType.RATE_LIMIT.value:
        return S2RateLimitError(
            message=message,
            details=details,
            status_code=429,
            endpoint=endpoint,
            method=method,
            params=params,
            json_body=json_body,
            base_url=base_url,
            response_text=details.get("response"),
            retry_after=details.get("retry_after"),
            authenticated=bool(details.get("authenticated", False)),
        )

    if error_type == ErrorType.TIMEOUT.value:
        return S2TimeoutError(
            message=message,
            details=details,
            endpoint=endpoint,
            method=method,
            timeout_seconds=Config.TIMEOUT,
        )

    status_code = details.get("status_code")
    exc_cls = S2NotFoundError if status_code == 404 else S2ApiError
    return exc_cls(
        message=message,
        details=details,
        status_code=status_code,
        endpoint=endpoint,
        method=method,
        params=params,
        json_body=json_body,
        base_url=base_url,
        response_text=details.get("response"),
    )


class S2Transport:
    async def request_json(
        self,
        endpoint: str,
        *,
        params: Optional[dict[str, Any]] = None,
        api_key_override: Optional[str] = None,
        method: str = "GET",
        json: Any = None,
        base_url: Optional[str] = None,
    ) -> Any:
        api_key = _normalize_key(api_key_override) or _normalize_key(get_api_key())
        authenticated = bool(api_key)

        await rate_limiter.acquire(endpoint, authenticated=authenticated, base_url=base_url)

        if api_key:
            headers = {"x-api-key": api_key}
        else:
            headers = {}
            logger.debug("Not sending x-api-key header (no valid API key available)")

        headers.setdefault(
            "User-Agent",
            "semantic-scholar-mcp/1.0 (+https://github.com/zongmin-yu/semantic-scholar-fastmcp-mcp-server)",
        )

        if endpoint.startswith("http://") or endpoint.startswith("https://"):
            url = endpoint
        else:
            url = f"{base_url or Config.BASE_URL}{endpoint}"

        try:
            client = await initialize_client()
            logger.debug(
                "Semantic Scholar request: method=%s url=%s params=%s headers=%s",
                method,
                url,
                params,
                _redact_headers(headers),
            )
            response = await client.request(method.upper(), url, params=params, headers=headers, json=json)
            response.raise_for_status()
            return response.json()
        except httpx.HTTPStatusError as exc:
            try:
                logger.error("HTTP error %s for %s: %s", exc.response.status_code, url, exc.response.text)
                logger.error("Request headers: %s", _redact_headers(headers))
                logger.error("Request params: %s", params)
            except Exception:
                logger.exception("Failed to log request details")

            status_code = exc.response.status_code
            response_text = exc.response.text
            if status_code == 429:
                raise S2RateLimitError(
                    message="Rate limit exceeded. Consider using an API key for higher limits.",
                    details={},
                    status_code=429,
                    endpoint=endpoint,
                    method=method,
                    params=params,
                    json_body=json,
                    base_url=base_url,
                    response_text=response_text,
                    retry_after=exc.response.headers.get("retry-after"),
                    authenticated=authenticated,
                ) from exc

            error_cls = S2NotFoundError if status_code == 404 else S2ApiError
            raise error_cls(
                message=f"HTTP error: {status_code}",
                details={},
                status_code=status_code,
                endpoint=endpoint,
                method=method,
                params=params,
                json_body=json,
                base_url=base_url,
                response_text=response_text,
            ) from exc
        except httpx.TimeoutException as exc:
            logger.error("Request timeout for %s: %s", endpoint, str(exc))
            raise S2TimeoutError(
                message=f"Request timed out after {Config.TIMEOUT} seconds",
                details={},
                endpoint=endpoint,
                method=method,
                timeout_seconds=Config.TIMEOUT,
            ) from exc
        except Exception as exc:
            logger.error("Unexpected error for %s: %s", endpoint, str(exc))
            raise S2ApiError(
                message=str(exc),
                details={},
                endpoint=endpoint,
                method=method,
                params=params,
                json_body=json,
                base_url=base_url,
            ) from exc


class MakeRequestCompatTransport:
    def __init__(self, make_request_callable) -> None:
        self._make_request = make_request_callable

    async def request_json(
        self,
        endpoint: str,
        *,
        params: Optional[dict[str, Any]] = None,
        api_key_override: Optional[str] = None,
        method: str = "GET",
        json: Any = None,
        base_url: Optional[str] = None,
    ) -> Any:
        result = await self._make_request(
            endpoint,
            params=params,
            api_key_override=api_key_override,
            method=method,
            json=json,
            base_url=base_url,
        )
        if isinstance(result, dict) and "error" in result:
            raise error_dict_to_exception(
                result,
                endpoint=endpoint,
                method=method,
                params=params,
                json_body=json,
                base_url=base_url,
            )
        return result


default_transport = S2Transport()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from ..config import (
    AuthorDetailFields,
    CitationReferenceFields,
    Config,
    PaperFields,
    VALID_FIELDS_OF_STUDY,
    VALID_PUBLICATION_TYPES,
    VALID_RECOMMENDATION_POOLS,
)
from .exceptions import S2ValidationError

VALID_BULK_SORT_FIELDS = ["paperId", "publicationDate", "citationCount"]
VALID_BULK_SORT_ORDERS = ["asc", "desc"]


class RequestModel:
    method: str = "GET"
    base_url: Optional[str] = None

    @property
    def endpoint(self) -> str:
        raise NotImplementedError

    def to_params(self) -> dict[str, Any]:
        raise NotImplementedError

    def to_json(self) -> Any:
        return None


def _raise_validation(message: str, details: Optional[dict[str, Any]] = None, field: Optional[str] = None) -> None:
    raise S2ValidationError(message=message, details=details or {}, field=field)


def _validate_fields(fields: list[str], valid_fields: set[str]) -> None:
    invalid_fields = set(fields) - valid_fields
    if invalid_fields:
        _raise_validation(
            f"Invalid fields: {', '.join(invalid_fields)}",
            {"valid_fields": list(valid_fields)},
            field="fields",
        )


def _validate_csv_fields(fields: str, valid_fields: set[str]) -> None:
    field_list = fields.split(",")
    invalid_fields = set(field_list) - valid_fields
    if invalid_fields:
        _raise_validation(
            f"Invalid fields: {', '.join(invalid_fields)}",
            {"valid_fields": list(valid_fields)},
            field="fields",
        )


@dataclass(slots=True)
class PaperRelevanceSearchRequest(RequestModel):
    query: str
    fields: Optional[list[str]] = None
    publication_types: Optional[list[str]] = None
    open_access_pdf: bool = False
    min_citation_count: Optional[int] = None
    year: Optional[str] = None
    venue: Optional[list[str]] = None
    fields_of_study: Optional[list[str]] = None
    offset: int = 0
    limit: int = 10

    @property
    def endpoint(self) -> str:
        return "/paper/search"

    def __post_init__(self) -> None:
        if not self.query.strip():
            _raise_validation("Query string cannot be empty", field="query")
        if self.fields is None:
            self.fields = list(PaperFields.DEFAULT)
        else:
            _validate_fields(self.fields, PaperFields.VALID_FIELDS)
        self.limit = min(self.limit, 100)

    def to_params(self) -> dict[str, Any]:
        params = {
            "query": self.query,
            "offset": self.offset,
            "limit": self.limit,
            "fields": ",".join(self.fields or []),
        }
        if self.publication_types:
            params["publicationTypes"] = ",".join(self.publication_types)
        if self.open_access_pdf:
            params["openAccessPdf"] = "true"
        if self.min_citation_count is not None:
            params["minCitationCount"] = self.min_citation_count
        if self.year:
            params["year"] = self.year
        if self.venue:
            params["venue"] = ",".join(self.venue)
        if self.fields_of_study:
            params["fieldsOfStudy"] = ",".join(self.fields_of_study)
        return params


@dataclass(slots=True)
class PaperBulkSearchRequest(RequestModel):
    query: Optional[str] = None
    token: Optional[str] = None
    fields: Optional[list[str]] = None
    sort: Optional[str] = None
    publication_types: Optional[list[str]] = None
    open_access_pdf: bool = False
    min_citation_count: Optional[int] = None
    publication_date_or_year: Optional[str] = None
    year: Optional[str] = None
    venue: Optional[list[str]] = None
    fields_of_study: Optional[list[str]] = None

    @property
    def endpoint(self) -> str:
        return "/paper/search/bulk"

    def __post_init__(self) -> None:
        if self.fields:
            _validate_fields(self.fields, PaperFields.VALID_FIELDS)
        if self.sort:
            try:
                field, order = self.sort.split(":")
            except ValueError as exc:
                _raise_validation("Sort must be in format 'field:order'", field="sort")
                raise AssertionError("unreachable") from exc
            if field not in VALID_BULK_SORT_FIELDS:
                _raise_validation(
                    "Invalid sort field. Must be one of: paperId, publicationDate, citationCount",
                    field="sort",
                )
            if order not in VALID_BULK_SORT_ORDERS:
                _raise_validation(
                    "Invalid sort order. Must be one of: asc, desc",
                    field="sort",
                )
        if self.publication_types:
            invalid_types = set(self.publication_types) - VALID_PUBLICATION_TYPES
            if invalid_types:
                _raise_validation(
                    f"Invalid publication types: {', '.join(invalid_types)}",
                    {"valid_types": list(VALID_PUBLICATION_TYPES)},
                    field="publication_types",
                )
        if self.min_citation_count is not None and self.min_citation_count < 0:
            _raise_validation("Minimum citation count cannot be negative", field="min_citation_count")
        if self.fields_of_study:
            invalid_fields = set(self.fields_of_study) - VALID_FIELDS_OF_STUDY
            if invalid_fields:
                _raise_validation(
                    f"Invalid fields of study: {', '.join(invalid_fields)}",
                    {"valid_fields": list(VALID_FIELDS_OF_STUDY)},
                    field="fields_of_study",
                )

    def to_params(self) -> dict[str, Any]:
        params: dict[str, Any] = {}
        if self.query:
            params["query"] = self.query.strip()
        if self.token:
            params["token"] = self.token
        if self.fields:
            params["fields"] = ",".join(self.fields)
        if self.sort:
            params["sort"] = self.sort
        if self.publication_types:
            params["publicationTypes"] = ",".join(self.publication_types)
        if self.open_access_pdf:
            params["openAccessPdf"] = "true"
        if self.min_citation_count is not None:
            params["minCitationCount"] = str(self.min_citation_count)
        if self.publication_date_or_year:
            params["publicationDateOrYear"] = self.publication_date_or_year
        elif self.year:
            params["year"] = self.year
        if self.venue:
            params["venue"] = ",".join(self.venue)
        if self.fields_of_study:
            params["fieldsOfStudy"] = ",".join(self.fields_of_study)
        return params


@dataclass(slots=True)
class PaperTitleSearchRequest(RequestModel):
    query: str
    fields: Optional[list[str]] = None
    publication_types: Optional[list[str]] = None
    open_access_pdf: bool = False
    min_citation_count: Optional[int] = None
    year: Optional[str] = None
    venue: Optional[list[str]] = None
    fields_of_study: Optional[list[str]] = None

    @property
    def endpoint(self) -> str:
        return "/paper/search/match"

    def __post_init__(self) -> None:
        if not self.query.strip():
            _raise_validation("Query string cannot be empty", field="query")
        if self.fields is None:
            self.fields = list(PaperFields.DEFAULT)
        else:
            _validate_fields(self.fields, PaperFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {"query": self.query, "fields": ",".join(self.fields or [])}
        if self.publication_types:
            params["publicationTypes"] = ",".join(self.publication_types)
        if self.open_access_pdf:
            params["openAccessPdf"] = "true"
        if self.min_citation_count is not None:
            params["minCitationCount"] = str(self.min_citation_count)
        if self.year:
            params["year"] = self.year
        if self.venue:
            params["venue"] = ",".join(self.venue)
        if self.fields_of_study:
            params["fieldsOfStudy"] = ",".join(self.fields_of_study)
        return params


@dataclass(slots=True)
class PaperDetailsRequest(RequestModel):
    paper_id: str
    fields: Optional[list[str]] = None

    @property
    def endpoint(self) -> str:
        return f"/paper/{self.paper_id}"

    def __post_init__(self) -> None:
        if not self.paper_id.strip():
            _raise_validation("Paper ID cannot be empty", field="paper_id")

    def to_params(self) -> dict[str, Any]:
        params = {}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class PaperBatchDetailsRequest(RequestModel):
    paper_ids: list[str]
    fields: Optional[str] = None
    method: str = "POST"

    @property
    def endpoint(self) -> str:
        return "/paper/batch"

    def __post_init__(self) -> None:
        if not self.paper_ids:
            _raise_validation("Paper IDs list cannot be empty", field="paper_ids")
        if len(self.paper_ids) > 500:
            _raise_validation(
                "Cannot process more than 500 paper IDs at once",
                {"max_papers": 500, "received": len(self.paper_ids)},
                field="paper_ids",
            )
        if self.fields:
            _validate_csv_fields(self.fields, PaperFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {}
        if self.fields:
            params["fields"] = self.fields
        return params

    def to_json(self) -> Any:
        return {"ids": self.paper_ids}


@dataclass(slots=True)
class PaperAuthorsRequest(RequestModel):
    paper_id: str
    fields: Optional[list[str]] = None
    offset: int = 0
    limit: int = 100

    @property
    def endpoint(self) -> str:
        return f"/paper/{self.paper_id}/authors"

    def __post_init__(self) -> None:
        if not self.paper_id.strip():
            _raise_validation("Paper ID cannot be empty", field="paper_id")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")
        if self.fields:
            _validate_fields(self.fields, AuthorDetailFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {"offset": self.offset, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class PaperCitationsRequest(RequestModel):
    paper_id: str
    fields: Optional[list[str]] = None
    offset: int = 0
    limit: int = 100

    @property
    def endpoint(self) -> str:
        return f"/paper/{self.paper_id}/citations"

    def __post_init__(self) -> None:
        if not self.paper_id.strip():
            _raise_validation("Paper ID cannot be empty", field="paper_id")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")
        if self.fields:
            _validate_fields(self.fields, CitationReferenceFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {"offset": self.offset, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class PaperReferencesRequest(RequestModel):
    paper_id: str
    fields: Optional[list[str]] = None
    offset: int = 0
    limit: int = 100

    @property
    def endpoint(self) -> str:
        return f"/paper/{self.paper_id}/references"

    def __post_init__(self) -> None:
        if not self.paper_id.strip():
            _raise_validation("Paper ID cannot be empty", field="paper_id")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")
        if self.fields:
            _validate_fields(self.fields, CitationReferenceFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {"offset": self.offset, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class PaperAutocompleteRequest(RequestModel):
    query: str

    @property
    def endpoint(self) -> str:
        return "/paper/autocomplete"

    def __post_init__(self) -> None:
        if not self.query.strip():
            _raise_validation("Query string cannot be empty", field="query")

    def to_params(self) -> dict[str, Any]:
        return {"query": self.query[:100]}


@dataclass(slots=True)
class SnippetSearchRequest(RequestModel):
    query: str
    fields: Optional[list[str]] = None
    limit: int = 10
    paper_ids: Optional[list[str]] = None
    authors: Optional[list[str]] = None
    min_citation_count: Optional[int] = None
    inserted_before: Optional[str] = None
    publication_date_or_year: Optional[str] = None
    year: Optional[str] = None
    venue: Optional[list[str]] = None
    fields_of_study: Optional[list[str]] = None

    @property
    def endpoint(self) -> str:
        return "/snippet/search"

    def __post_init__(self) -> None:
        if not self.query.strip():
            _raise_validation("Query string cannot be empty", field="query")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")
        if self.limit < 1:
            _raise_validation("Limit must be at least 1", {"min_limit": 1}, field="limit")
        if self.authors and len(self.authors) > 10:
            _raise_validation("Cannot filter by more than 10 authors", {"max_authors": 10}, field="authors")
        if self.paper_ids and len(self.paper_ids) > 100:
            _raise_validation(
                "Cannot filter by more than 100 paper IDs",
                {"max_paper_ids": 100},
                field="paper_ids",
            )

    def to_params(self) -> dict[str, Any]:
        params = {"query": self.query, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        if self.paper_ids:
            params["paperIds"] = ",".join(self.paper_ids)
        if self.authors:
            params["authors"] = ",".join(self.authors)
        if self.min_citation_count is not None:
            params["minCitationCount"] = self.min_citation_count
        if self.inserted_before:
            params["insertedBefore"] = self.inserted_before
        if self.publication_date_or_year:
            params["publicationDateOrYear"] = self.publication_date_or_year
        if self.year:
            params["year"] = self.year
        if self.venue:
            params["venue"] = ",".join(self.venue)
        if self.fields_of_study:
            params["fieldsOfStudy"] = ",".join(self.fields_of_study)
        return params


@dataclass(slots=True)
class AuthorSearchRequest(RequestModel):
    query: str
    fields: Optional[list[str]] = None
    offset: int = 0
    limit: int = 100

    @property
    def endpoint(self) -> str:
        return "/author/search"

    def __post_init__(self) -> None:
        if not self.query.strip():
            _raise_validation("Query string cannot be empty", field="query")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")
        if self.fields:
            _validate_fields(self.fields, AuthorDetailFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {"query": self.query, "offset": self.offset, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class AuthorDetailsRequest(RequestModel):
    author_id: str
    fields: Optional[list[str]] = None

    @property
    def endpoint(self) -> str:
        return f"/author/{self.author_id}"

    def __post_init__(self) -> None:
        if not self.author_id.strip():
            _raise_validation("Author ID cannot be empty", field="author_id")
        if self.fields:
            _validate_fields(self.fields, AuthorDetailFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class AuthorPapersRequest(RequestModel):
    author_id: str
    fields: Optional[list[str]] = None
    offset: int = 0
    limit: int = 100

    @property
    def endpoint(self) -> str:
        return f"/author/{self.author_id}/papers"

    def __post_init__(self) -> None:
        if not self.author_id.strip():
            _raise_validation("Author ID cannot be empty", field="author_id")
        if self.limit > 1000:
            _raise_validation("Limit cannot exceed 1000", {"max_limit": 1000}, field="limit")

    def to_params(self) -> dict[str, Any]:
        params = {"offset": self.offset, "limit": self.limit}
        if self.fields:
            params["fields"] = ",".join(self.fields)
        return params


@dataclass(slots=True)
class AuthorBatchDetailsRequest(RequestModel):
    author_ids: list[str]
    fields: Optional[str] = None
    method: str = "POST"

    @property
    def endpoint(self) -> str:
        return "/author/batch"

    def __post_init__(self) -> None:
        if not self.author_ids:
            _raise_validation("Author IDs list cannot be empty", field="author_ids")
        if len(self.author_ids) > 1000:
            _raise_validation(
                "Cannot process more than 1000 author IDs at once",
                {"max_authors": 1000, "received": len(self.author_ids)},
                field="author_ids",
            )
        if self.fields:
            _validate_csv_fields(self.fields, AuthorDetailFields.VALID_FIELDS)

    def to_params(self) -> dict[str, Any]:
        params = {}
        if self.fields:
            params["fields"] = self.fields
        return params

    def to_json(self) -> Any:
        return {"ids": self.author_ids}


@dataclass(slots=True)
class PaperRecommendationsSingleRequest(RequestModel):
    paper_id: str
    fields: Optional[str] = None
    limit: int = 100
    from_pool: str = "recent"
    base_url: str = Config.RECOMMENDATIONS_BASE_URL

    @property
    def endpoint(self) -> str:
        return f"/papers/forpaper/{self.paper_id}"

    def __post_init__(self) -> None:
        if self.limit > 500:
            _raise_validation(
                "Cannot request more than 500 recommendations",
                {"max_limit": 500, "requested": self.limit},
                field="limit",
            )
        if self.from_pool not in VALID_RECOMMENDATION_POOLS:
            _raise_validation(
                "Invalid paper pool specified",
                {"valid_pools": list(VALID_RECOMMENDATION_POOLS)},
                field="from_pool",
            )

    def to_params(self) -> dict[str, Any]:
        params = {"limit": self.limit, "from": self.from_pool}
        if self.fields:
            params["fields"] = self.fields
        return params


@dataclass(slots=True)
class PaperRecommendationsMultiRequest(RequestModel):
    positive_paper_ids: list[str]
    negative_paper_ids: Optional[list[str]] = None
    fields: Optional[str] = None
    limit: int = 100
    method: str = "POST"
    base_url: str = Config.RECOMMENDATIONS_BASE_URL

    @property
    def endpoint(self) -> str:
        return "/papers"

    def __post_init__(self) -> None:
        if not self.positive_paper_ids:
            _raise_validation("Must provide at least one positive paper ID", field="positive_paper_ids")
        if self.limit > 500:
            _raise_validation(
                "Cannot request more than 500 recommendations",
                {"max_limit": 500, "requested": self.limit},
                field="limit",
            )

    def to_params(self) -> dict[str, Any]:
        params = {"limit": self.limit}
        if self.fields:
            params["fields"] = self.fields
        return params

    def to_json(self) -> Any:
        return {
            "positivePaperIds": self.positive_paper_ids,
            "negativePaperIds": self.negative_paper_ids or [],
        }



"""
Configuration for the Semantic Scholar API Server.
"""

from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Tuple, Any
import os

# Rate Limiting Configuration
@dataclass
class RateLimitConfig:
    # Define rate limits (requests, seconds)
    SEARCH_LIMIT = (1, 1)  # 1 request per 1 second
    BATCH_LIMIT = (1, 1)   # 1 request per 1 second
    RECOMMENDATIONS_LIMIT = (1, 1)  # 1 request per 1 second
    DEFAULT_LIMIT = (10, 1)  # 10 requests per 1 second
    UNAUTHENTICATED_LIMIT = (100, 300)  # 100 requests per 5 minutes
    
    # Endpoints categorization
    # These endpoints have stricter rate limits due to their computational intensity
    # and to prevent abuse of the recommendation system
    RESTRICTED_ENDPOINTS = [
        "/author/batch",    # Batch operations are expensive
        "/author/search",   # Search operations are computationally intensive
        "/paper/batch",     # Batch operations are expensive
        "/paper/search",    # Search operations are computationally intensive
        "/recommendations"  # Recommendation generation is resource-intensive
    ]

# Error Types
class ErrorType(Enum):
    RATE_LIMIT = "rate_limit"
    API_ERROR = "api_error"
    VALIDATION = "validation"
    TIMEOUT = "timeout"


VALID_PUBLICATION_TYPES = {
    "Review",
    "JournalArticle",
    "CaseReport",
    "ClinicalTrial",
    "Conference",
    "Dataset",
    "Editorial",
    "LettersAndComments",
    "MetaAnalysis",
    "News",
    "Study",
    "Book",
    "BookSection",
}

VALID_FIELDS_OF_STUDY = {
    "Computer Science",
    "Medicine",
    "Chemistry",
    "Biology",
    "Materials Science",
    "Physics",
    "Geology",
    "Psychology",
    "Art",
    "History",
    "Geography",
    "Sociology",
    "Business",
    "Political Science",
    "Economics",
    "Philosophy",
    "Mathematics",
    "Engineering",
    "Environmental Science",
    "Agricultural and Food Sciences",
    "Education",
    "Law",
    "Linguistics",
}

VALID_RECOMMENDATION_POOLS = ["recent", "all-cs"]

# Field Constants
class PaperFields:
    DEFAULT = ["title", "abstract", "year", "citationCount", "authors", "url"]
    DETAILED = DEFAULT + ["references", "citations", "venue", "influentialCitationCount"]
    MINIMAL = ["title", "year", "authors"]
    SEARCH = ["paperId", "title", "year", "citationCount"]
    
    # Valid fields from API documentation
    VALID_FIELDS = {
        "abstract",
        "authors",
        "citationCount",
        "citationStyles",
        "citations",
        "corpusId",
        "embedding",
        "externalIds",
        "fieldsOfStudy",
        "influentialCitationCount",
        "isOpenAccess",
        "journal",
        "openAccessPdf",
        "paperId",
        "publicationDate",
        "publicationTypes",
        "publicationVenue",
        "referenceCount",
        "references",
        "s2FieldsOfStudy",
        "textAvailability",
        "title",
        "tldr",
        "url",
        "venue",
        "year"
    }

class AuthorDetailFields:
    """Common field combinations for author details"""
    
    # Basic author information
    BASIC = ["name", "url", "affiliations"]
    
    # Author's papers information
    PAPERS_BASIC = ["papers"]  # Returns paperId and title
    PAPERS_DETAILED = [
        "papers.year",
        "papers.authors",
        "papers.abstract",
        "papers.venue",
        "papers.url"
    ]
    
    # Complete author profile
    COMPLETE = BASIC + ["papers", "papers.year", "papers.authors", "papers.venue"]
    
    # Citation metrics
    METRICS = ["citationCount", "hIndex", "paperCount"]

    # Valid fields for author details
    VALID_FIELDS = {
        "affiliations",
        "authorId",
        "citationCount",
        "externalIds",
        "hIndex",
        "homepage",
        "name",
        "paperCount",
        "papers",
        "papers.abstract",
        "papers.authors",
        "papers.citationCount",
        "papers.citationStyles",
        "papers.corpusId",
        "papers.externalIds",
        "papers.fieldsOfStudy",
        "papers.influentialCitationCount",
        "papers.isOpenAccess",
        "papers.journal",
        "papers.openAccessPdf",
        "papers.paperId",
        "papers.publicationDate",
        "papers.publicationTypes",
        "papers.publicationVenue",
        "papers.referenceCount",
        "papers.s2FieldsOfStudy",
        "papers.title",
        "papers.url",
        "papers.venue",
        "papers.year",
        "url",
    }

class PaperDetailFields:
    """Common field combinations for paper details"""
    
    # Basic paper information
    BASIC = ["title", "abstract", "year", "venue"]
    
    # Author information
    AUTHOR_BASIC = ["authors"]
    AUTHOR_DETAILED = ["authors.url", "authors.paperCount", "authors.citationCount"]
    
    # Citation information
    CITATION_BASIC = ["citations", "references"]
    CITATION_DETAILED = ["citations.title", "citations.abstract", "citations.year",
                        "references.title", "references.abstract", "references.year"]
    
    # Full paper details
    COMPLETE = BASIC + AUTHOR_BASIC + CITATION_BASIC + ["url", "fieldsOfStudy", 
                                                       "publicationVenue", "publicationTypes"]

class CitationReferenceFields:
    """Common field combinations for citation and reference queries"""
    
    # Basic information
    BASIC = ["title"]
    
    # Citation/Reference context
    CONTEXT = ["contexts", "intents", "isInfluential"]
    
    # Paper details
    DETAILED = ["title", "abstract", "authors", "year", "venue"]
    
    # Full information
    COMPLETE = CONTEXT + DETAILED

    # Valid fields for citation/reference queries
    VALID_FIELDS = {
        "abstract",
        "authors",
        "citationCount",
        "citationStyles",
        "contexts",
        "contextsWithIntent",
        "corpusId",
        "externalIds",
        "fieldsOfStudy",
        "influentialCitationCount",
        "intents",
        "isInfluential",
        "isOpenAccess",
        "journal",
        "openAccessPdf",
        "paperId",
        "publicationDate",
        "publicationTypes",
        "publicationVenue",
        "referenceCount",
        "s2FieldsOfStudy",
        "title",
        "url",
        "venue",
        "year",
    }

# Configuration
class Config:
    # API Configuration
    API_VERSION = "v1"
    BASE_URL = f"https://api.semanticscholar.org/graph/{API_VERSION}"
    RECOMMENDATIONS_BASE_URL = "https://api.semanticscholar.org/recommendations/v1"
    TIMEOUT = int(os.getenv("SEMANTIC_SCHOLAR_TIMEOUT", "30"))  # seconds
    
    # Request Limits
    MAX_BATCH_SIZE = 100
    MAX_RESULTS_PER_PAGE = 100
    DEFAULT_PAGE_SIZE = 10
    MAX_BATCHES = 5
    
    # Fields Configuration
    DEFAULT_FIELDS = PaperFields.DEFAULT
    
    # Feature Flags
    ENABLE_CACHING = False
    DEBUG_MODE = False
    
    # Search Configuration
    SEARCH_TYPES = {
        "comprehensive": {
            "description": "Balanced search considering relevance and impact",
            "min_citations": None,
            "ranking_strategy": "balanced"
        },
        "influential": {
            "description": "Focus on highly-cited and influential papers",
            "min_citations": 50,
            "ranking_strategy": "citations"
        },
        "latest": {
            "description": "Focus on recent papers with impact",
            "min_citations": None,
            "ranking_strategy": "recency"
        }
    } 



import pytest

import semantic_scholar.api.recommendations as recommendations_api
from semantic_scholar.config import Config


pytestmark = pytest.mark.asyncio


def assert_single_call(
    mock_make_request,
    *,
    endpoint,
    params,
    api_key_override=None,
    method="GET",
    json=None,
    base_url=None,
):
    assert mock_make_request.calls == [
        {
            "endpoint": endpoint,
            "params": params,
            "api_key_override": api_key_override,
            "method": method,
            "json": json,
            "base_url": base_url,
        }
    ]


def assert_validation_error(result, message, details=None):
    assert result == {
        "error": {
            "type": "validation",
            "message": message,
            "details": details or {},
        }
    }


async def test_get_paper_recommendations_single_happy_path(mock_make_request):
    payload = {"recommendedPapers": [{"paperId": "p1"}]}
    mock_make_request.install(recommendations_api).queue_responses(payload)

    result = await recommendations_api.get_paper_recommendations_single.fn(
        None,
        paper_id="204e3073870fae3d05bcbc2f6a8e263d9b72e776",
        fields="title,year",
        limit=25,
        from_pool="all-cs",
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/papers/forpaper/204e3073870fae3d05bcbc2f6a8e263d9b72e776",
        params={"limit": 25, "from": "all-cs", "fields": "title,year"},
        base_url=Config.RECOMMENDATIONS_BASE_URL,
    )


async def test_get_paper_recommendations_single_invalid_pool_validation(mock_make_request):
    result = await recommendations_api.get_paper_recommendations_single.fn(
        None,
        paper_id="204e3073870fae3d05bcbc2f6a8e263d9b72e776",
        from_pool="legacy",
    )

    assert_validation_error(
        result,
        "Invalid paper pool specified",
        {"valid_pools": ["recent", "all-cs"]},
    )
    assert mock_make_request.calls == []


async def test_get_paper_recommendations_single_404_maps_to_paper_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(recommendations_api).queue_responses(mock_error_response(status_code=404))

    result = await recommendations_api.get_paper_recommendations_single.fn(
        None,
        paper_id="204e3073870fae3d05bcbc2f6a8e263d9b72e776",
        fields="title,year",
        limit=25,
        from_pool="all-cs",
    )

    assert_validation_error(
        result,
        "Paper not found",
        {"paper_id": "204e3073870fae3d05bcbc2f6a8e263d9b72e776"},
    )
    assert_single_call(
        mock_make_request,
        endpoint="/papers/forpaper/204e3073870fae3d05bcbc2f6a8e263d9b72e776",
        params={"limit": 25, "from": "all-cs", "fields": "title,year"},
        base_url=Config.RECOMMENDATIONS_BASE_URL,
    )


async def test_get_paper_recommendations_multi_happy_path(mock_make_request):
    payload = {"recommendedPapers": [{"paperId": "p4"}]}
    mock_make_request.install(recommendations_api).queue_responses(payload)

    result = await recommendations_api.get_paper_recommendations_multi.fn(
        None,
        positive_paper_ids=["p1", "p2"],
        negative_paper_ids=["p3"],
        fields="title,year",
        limit=25,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/papers",
        params={"limit": 25, "fields": "title,year"},
        method="POST",
        json={"positivePaperIds": ["p1", "p2"], "negativePaperIds": ["p3"]},
        base_url=Config.RECOMMENDATIONS_BASE_URL,
    )


async def test_get_paper_recommendations_multi_requires_positive_ids(mock_make_request):
    result = await recommendations_api.get_paper_recommendations_multi.fn(None, positive_paper_ids=[])

    assert_validation_error(result, "Must provide at least one positive paper ID")
    assert mock_make_request.calls == []


async def test_get_paper_recommendations_multi_404_maps_to_missing_input_papers(
    mock_make_request,
    mock_error_response,
):
    mock_make_request.install(recommendations_api).queue_responses(mock_error_response(status_code=404))

    result = await recommendations_api.get_paper_recommendations_multi.fn(
        None,
        positive_paper_ids=["p1", "p2"],
        negative_paper_ids=["p3"],
        fields="title,year",
        limit=25,
    )

    assert_validation_error(
        result,
        "One or more input papers not found",
        {"positive_ids": ["p1", "p2"], "negative_ids": ["p3"]},
    )
    assert_single_call(
        mock_make_request,
        endpoint="/papers",
        params={"limit": 25, "fields": "title,year"},
        method="POST",
        json={"positivePaperIds": ["p1", "p2"], "negativePaperIds": ["p3"]},
        base_url=Config.RECOMMENDATIONS_BASE_URL,
    )



#!/usr/bin/env python3
"""
Entry point script for the Semantic Scholar API Server.

Available tools:
- paper_relevance_search
- paper_bulk_search
- paper_title_search
- paper_details
- paper_batch_details
- paper_authors
- paper_citations
- paper_references
- author_search
- author_details
- author_papers
- author_batch_details
- get_paper_recommendations_single
- get_paper_recommendations_multi
"""

# Import the mcp instance from centralized location
from semantic_scholar.mcp import mcp
# Import the main function from server
from semantic_scholar.server import main

# Import all API modules to ensure tools are registered
from semantic_scholar.api import papers, authors, recommendations

if __name__ == "__main__":
    main() 


import pytest

from semantic_scholar.config import RateLimitConfig
from semantic_scholar.utils.http import RateLimiter


@pytest.mark.asyncio
async def test_rate_limiter_uses_shared_bucket_for_dynamic_ids(monkeypatch):
    monkeypatch.setattr(RateLimitConfig, "DEFAULT_LIMIT", (2, 10))

    now = 0.0

    def clock():
        return now

    async def sleeper(delay: float):
        nonlocal now
        now += delay

    rl = RateLimiter(clock=clock, sleeper=sleeper)

    await rl.acquire("/paper/1", authenticated=True)
    await rl.acquire("/paper/2", authenticated=True)
    assert now == 0.0

    await rl.acquire("/paper/3", authenticated=True)
    assert now == 10.0


@pytest.mark.asyncio
async def test_rate_limiter_uses_unauthenticated_limits(monkeypatch):
    monkeypatch.setattr(RateLimitConfig, "DEFAULT_LIMIT", (100, 1))
    monkeypatch.setattr(RateLimitConfig, "UNAUTHENTICATED_LIMIT", (1, 5))

    now = 0.0

    def clock():
        return now

    async def sleeper(delay: float):
        nonlocal now
        now += delay

    rl = RateLimiter(clock=clock, sleeper=sleeper)

    await rl.acquire("/paper/search", authenticated=False)
    await rl.acquire("/paper/search", authenticated=False)
    assert now == 5.0


@pytest.mark.asyncio
async def test_rate_limiter_detects_recommendations_base_url(monkeypatch):
    monkeypatch.setattr(RateLimitConfig, "RECOMMENDATIONS_LIMIT", (1, 3))

    now = 0.0

    def clock():
        return now

    async def sleeper(delay: float):
        nonlocal now
        now += delay

    rl = RateLimiter(clock=clock, sleeper=sleeper)

    await rl.acquire(
        "/papers/forpaper/abc",
        authenticated=True,
        base_url="https://api.semanticscholar.org/recommendations/v1",
    )
    await rl.acquire(
        "/papers/forpaper/def",
        authenticated=True,
        base_url="https://api.semanticscholar.org/recommendations/v1",
    )
    assert now == 3.0




"""
HTTP Bridge for the Semantic Scholar MCP Server.
"""

from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import List, Optional

from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel

from .config import AuthorDetailFields, Config, PaperFields
from .core.client import S2Client, make_compat_client
from .core.exceptions import S2Error
from .core.requests import (
    AuthorBatchDetailsRequest,
    AuthorDetailsRequest,
    AuthorSearchRequest,
    PaperBatchDetailsRequest,
    PaperDetailsRequest,
    PaperRelevanceSearchRequest,
    RequestModel,
)
from .utils.errors import s2_exception_to_error_response
from .utils.http import cleanup_client, initialize_client, make_request


class IdList(BaseModel):
    ids: List[str]


@dataclass(slots=True)
class _BridgeRecommendationsRequest(RequestModel):
    paper_id: str
    fields: Optional[str] = None
    base_url: str = Config.RECOMMENDATIONS_BASE_URL

    @property
    def endpoint(self) -> str:
        return f"/papers/forpaper/{self.paper_id}"

    def to_params(self) -> dict[str, str]:
        params = {}
        if self.fields:
            params["fields"] = self.fields
        return params


@asynccontextmanager
async def lifespan(app: FastAPI):
    await initialize_client()
    try:
        yield
    finally:
        await cleanup_client()


app = FastAPI(title="Semantic Scholar Bridge", version="0.1", lifespan=lifespan)


def _client() -> S2Client:
    return make_compat_client(make_request)


def _bearer_token(request: Request) -> Optional[str]:
    auth = request.headers.get("authorization")
    if auth and auth.lower().startswith("bearer "):
        return auth.split(None, 1)[1]
    return None


def _bridge_error_response(exc: S2Error) -> JSONResponse:
    return JSONResponse(s2_exception_to_error_response(exc), status_code=200)


@app.get("/v1/paper/search")
async def paper_search(
    request: Request,
    q: str,
    fields: Optional[str] = None,
    offset: int = 0,
    limit: int = 10,
):
    token = _bearer_token(request)
    try:
        request_model = PaperRelevanceSearchRequest(
            query=q,
            fields=fields.split(",") if fields else Config.DEFAULT_FIELDS,
            offset=offset,
            limit=limit,
        )
        return await _client().search_papers(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.get("/v1/paper/{paper_id}")
async def paper_details(request: Request, paper_id: str, fields: Optional[str] = None):
    token = _bearer_token(request)
    try:
        request_model = PaperDetailsRequest(
            paper_id=paper_id,
            fields=fields.split(",") if fields else Config.DEFAULT_FIELDS,
        )
        return await _client().get_paper(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.post("/v1/paper/batch")
async def paper_batch(request: Request, batch: IdList, fields: Optional[str] = None):
    token = _bearer_token(request)
    try:
        request_model = PaperBatchDetailsRequest(
            paper_ids=batch.ids,
            fields=fields if fields else ",".join(Config.DEFAULT_FIELDS),
        )
        return await _client().batch_papers(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.get("/v1/author/search")
async def author_search(
    request: Request,
    q: str,
    fields: Optional[str] = None,
    offset: int = 0,
    limit: int = 10,
):
    token = _bearer_token(request)
    try:
        request_model = AuthorSearchRequest(
            query=q,
            fields=fields.split(",") if fields else AuthorDetailFields.BASIC,
            offset=offset,
            limit=limit,
        )
        return await _client().search_authors(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.get("/v1/author/{author_id}")
async def author_details(request: Request, author_id: str, fields: Optional[str] = None):
    token = _bearer_token(request)
    try:
        request_model = AuthorDetailsRequest(
            author_id=author_id,
            fields=fields.split(",") if fields else AuthorDetailFields.BASIC,
        )
        return await _client().get_author(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.post("/v1/author/batch")
async def author_batch(request: Request, batch: IdList, fields: Optional[str] = None):
    token = _bearer_token(request)
    try:
        request_model = AuthorBatchDetailsRequest(
            author_ids=batch.ids,
            fields=fields if fields else ",".join(AuthorDetailFields.BASIC),
        )
        return await _client().batch_authors(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)


@app.get("/v1/recommendations")
async def recommendations(request: Request, paper_id: Optional[str] = None, fields: Optional[str] = None):
    if not paper_id:
        raise HTTPException(status_code=400, detail="paper_id is required")

    token = _bearer_token(request)
    try:
        request_model = _BridgeRecommendationsRequest(
            paper_id=paper_id,
            fields=fields if fields else ",".join(PaperFields.DEFAULT),
        )
        return await _client().recommend_for_paper(request_model, api_key_override=token)
    except S2Error as exc:
        return _bridge_error_response(exc)



import pytest

import semantic_scholar.api.papers as papers_api


pytestmark = pytest.mark.asyncio


def assert_single_call(
    mock_make_request,
    *,
    endpoint,
    params,
    api_key_override=None,
    method="GET",
    json=None,
    base_url=None,
):
    assert mock_make_request.calls == [
        {
            "endpoint": endpoint,
            "params": params,
            "api_key_override": api_key_override,
            "method": method,
            "json": json,
            "base_url": base_url,
        }
    ]


def assert_validation_error(result, message, details=None):
    assert result == {
        "error": {
            "type": "validation",
            "message": message,
            "details": details or {},
        }
    }


async def test_paper_relevance_search_happy_path(mock_make_request):
    payload = {"total": 1, "offset": 0, "data": [{"paperId": "p1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_relevance_search.fn(None, query="attention", limit=5)

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search",
        params={
            "query": "attention",
            "offset": 0,
            "limit": 5,
            "fields": "title,abstract,year,citationCount,authors,url",
        },
    )


async def test_paper_relevance_search_empty_query(mock_make_request):
    result = await papers_api.paper_relevance_search.fn(None, query="   ")

    assert_validation_error(result, "Query string cannot be empty")
    assert mock_make_request.calls == []


async def test_paper_relevance_search_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(papers_api).queue_responses(error)

    result = await papers_api.paper_relevance_search.fn(None, query="attention", limit=5)

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search",
        params={
            "query": "attention",
            "offset": 0,
            "limit": 5,
            "fields": "title,abstract,year,citationCount,authors,url",
        },
    )


async def test_paper_bulk_search_happy_path(mock_make_request):
    payload = {"total": 1, "token": "next-2", "data": [{"paperId": "p1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_bulk_search.fn(
        None,
        query=" transformer ",
        token="next-1",
        fields=["paperId", "title"],
        sort="citationCount:desc",
        publication_types=["Conference"],
        open_access_pdf=True,
        min_citation_count=10,
        publication_date_or_year="2020-01-01:2024-12-31",
        venue=["NeurIPS"],
        fields_of_study=["Computer Science"],
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search/bulk",
        params={
            "query": "transformer",
            "token": "next-1",
            "fields": "paperId,title",
            "sort": "citationCount:desc",
            "publicationTypes": "Conference",
            "openAccessPdf": "true",
            "minCitationCount": "10",
            "publicationDateOrYear": "2020-01-01:2024-12-31",
            "venue": "NeurIPS",
            "fieldsOfStudy": "Computer Science",
        },
    )


async def test_paper_bulk_search_invalid_sort_field(mock_make_request):
    result = await papers_api.paper_bulk_search.fn(None, sort="year:desc")

    assert_validation_error(
        result,
        "Invalid sort field. Must be one of: paperId, publicationDate, citationCount",
    )
    assert mock_make_request.calls == []


async def test_paper_bulk_search_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(papers_api).queue_responses(error)

    result = await papers_api.paper_bulk_search.fn(
        None,
        query=" transformer ",
        token="next-1",
        fields=["paperId", "title"],
        sort="citationCount:desc",
        publication_types=["Conference"],
        open_access_pdf=True,
        min_citation_count=10,
        publication_date_or_year="2020-01-01:2024-12-31",
        venue=["NeurIPS"],
        fields_of_study=["Computer Science"],
    )

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search/bulk",
        params={
            "query": "transformer",
            "token": "next-1",
            "fields": "paperId,title",
            "sort": "citationCount:desc",
            "publicationTypes": "Conference",
            "openAccessPdf": "true",
            "minCitationCount": "10",
            "publicationDateOrYear": "2020-01-01:2024-12-31",
            "venue": "NeurIPS",
            "fieldsOfStudy": "Computer Science",
        },
    )


async def test_paper_title_search_happy_path(mock_make_request):
    payload = {"paperId": "p1", "title": "Attention is All You Need"}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_title_search.fn(None, query="Attention is All You Need")

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search/match",
        params={
            "query": "Attention is All You Need",
            "fields": "title,abstract,year,citationCount,authors,url",
        },
    )


async def test_paper_title_search_empty_query(mock_make_request):
    result = await papers_api.paper_title_search.fn(None, query="   ")

    assert_validation_error(result, "Query string cannot be empty")
    assert mock_make_request.calls == []


async def test_paper_title_search_404_maps_to_no_match(mock_make_request, mock_error_response):
    mock_make_request.install(papers_api).queue_responses(mock_error_response(status_code=404))

    result = await papers_api.paper_title_search.fn(None, query="Attention is All You Need")

    assert_validation_error(
        result,
        "No matching paper found",
        {"original_query": "Attention is All You Need"},
    )
    assert_single_call(
        mock_make_request,
        endpoint="/paper/search/match",
        params={
            "query": "Attention is All You Need",
            "fields": "title,abstract,year,citationCount,authors,url",
        },
    )


async def test_paper_details_happy_path(mock_make_request):
    payload = {"paperId": "CorpusId:215416146", "title": "Attention"}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_details.fn(None, paper_id="CorpusId:215416146", fields=["title", "year"])

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/CorpusId:215416146",
        params={"fields": "title,year"},
    )


async def test_paper_details_empty_paper_id(mock_make_request):
    result = await papers_api.paper_details.fn(None, paper_id="   ")

    assert_validation_error(result, "Paper ID cannot be empty")
    assert mock_make_request.calls == []


async def test_paper_details_404_maps_to_paper_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(papers_api).queue_responses(mock_error_response(status_code=404))

    result = await papers_api.paper_details.fn(None, paper_id="CorpusId:215416146", fields=["title", "year"])

    assert_validation_error(result, "Paper not found", {"paper_id": "CorpusId:215416146"})
    assert_single_call(
        mock_make_request,
        endpoint="/paper/CorpusId:215416146",
        params={"fields": "title,year"},
    )


async def test_paper_batch_details_happy_path(mock_make_request):
    payload = [{"paperId": "649def34f8be52c8b66281af98ae884c09aef38b"}, {"paperId": "ARXIV:2106.15928"}]
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_batch_details.fn(
        None,
        paper_ids=["649def34f8be52c8b66281af98ae884c09aef38b", "ARXIV:2106.15928"],
        fields="title,year",
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/batch",
        params={"fields": "title,year"},
        method="POST",
        json={"ids": ["649def34f8be52c8b66281af98ae884c09aef38b", "ARXIV:2106.15928"]},
    )


async def test_paper_batch_details_empty_ids_validation(mock_make_request):
    result = await papers_api.paper_batch_details.fn(None, paper_ids=[], fields="title,year")

    assert_validation_error(result, "Paper IDs list cannot be empty")
    assert mock_make_request.calls == []


async def test_paper_batch_details_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(papers_api).queue_responses(error)

    result = await papers_api.paper_batch_details.fn(
        None,
        paper_ids=["649def34f8be52c8b66281af98ae884c09aef38b", "ARXIV:2106.15928"],
        fields="title,year",
    )

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/paper/batch",
        params={"fields": "title,year"},
        method="POST",
        json={"ids": ["649def34f8be52c8b66281af98ae884c09aef38b", "ARXIV:2106.15928"]},
    )


async def test_paper_authors_happy_path(mock_make_request):
    payload = {"offset": 5, "next": 30, "data": [{"authorId": "a1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_authors.fn(
        None,
        paper_id="paper-123",
        fields=["name", "affiliations"],
        offset=5,
        limit=25,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/authors",
        params={"offset": 5, "limit": 25, "fields": "name,affiliations"},
    )


async def test_paper_authors_limit_validation(mock_make_request):
    result = await papers_api.paper_authors.fn(None, paper_id="paper-123", limit=1001)

    assert_validation_error(result, "Limit cannot exceed 1000", {"max_limit": 1000})
    assert mock_make_request.calls == []


async def test_paper_authors_404_maps_to_paper_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(papers_api).queue_responses(mock_error_response(status_code=404))

    result = await papers_api.paper_authors.fn(
        None,
        paper_id="paper-123",
        fields=["name", "affiliations"],
        offset=5,
        limit=25,
    )

    assert_validation_error(result, "Paper not found", {"paper_id": "paper-123"})
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/authors",
        params={"offset": 5, "limit": 25, "fields": "name,affiliations"},
    )


async def test_paper_citations_happy_path(mock_make_request):
    payload = {"offset": 0, "next": 10, "data": [{"paperId": "c1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_citations.fn(
        None,
        paper_id="paper-123",
        fields=["title", "contexts"],
        offset=0,
        limit=10,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/citations",
        params={"offset": 0, "limit": 10, "fields": "title,contexts"},
    )


async def test_paper_citations_invalid_field_validation(mock_make_request):
    result = await papers_api.paper_citations.fn(None, paper_id="paper-123", fields=["badField"])

    assert result["error"]["type"] == "validation"
    assert result["error"]["message"] == "Invalid fields: badField"
    assert mock_make_request.calls == []


async def test_paper_citations_404_maps_to_paper_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(papers_api).queue_responses(mock_error_response(status_code=404))

    result = await papers_api.paper_citations.fn(
        None,
        paper_id="paper-123",
        fields=["title", "contexts"],
        offset=0,
        limit=10,
    )

    assert_validation_error(result, "Paper not found", {"paper_id": "paper-123"})
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/citations",
        params={"offset": 0, "limit": 10, "fields": "title,contexts"},
    )


async def test_paper_references_happy_path(mock_make_request):
    payload = {"offset": 0, "next": 10, "data": [{"paperId": "r1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_references.fn(
        None,
        paper_id="paper-123",
        fields=["title", "contexts"],
        offset=0,
        limit=10,
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/references",
        params={"offset": 0, "limit": 10, "fields": "title,contexts"},
    )


async def test_paper_references_limit_validation(mock_make_request):
    result = await papers_api.paper_references.fn(None, paper_id="paper-123", limit=1001)

    assert_validation_error(result, "Limit cannot exceed 1000", {"max_limit": 1000})
    assert mock_make_request.calls == []


async def test_paper_references_404_maps_to_paper_not_found(mock_make_request, mock_error_response):
    mock_make_request.install(papers_api).queue_responses(mock_error_response(status_code=404))

    result = await papers_api.paper_references.fn(
        None,
        paper_id="paper-123",
        fields=["title", "contexts"],
        offset=0,
        limit=10,
    )

    assert_validation_error(result, "Paper not found", {"paper_id": "paper-123"})
    assert_single_call(
        mock_make_request,
        endpoint="/paper/paper-123/references",
        params={"offset": 0, "limit": 10, "fields": "title,contexts"},
    )


async def test_paper_autocomplete_happy_path(mock_make_request):
    long_query = "a" * 150
    payload = {"matches": [{"paperId": "p1"}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.paper_autocomplete.fn(None, query=long_query)

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/paper/autocomplete",
        params={"query": "a" * 100},
    )


async def test_paper_autocomplete_empty_query_validation(mock_make_request):
    result = await papers_api.paper_autocomplete.fn(None, query="   ")

    assert_validation_error(result, "Query string cannot be empty")
    assert mock_make_request.calls == []


async def test_paper_autocomplete_rate_limit_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(status_code=429)
    mock_make_request.install(papers_api).queue_responses(error)

    result = await papers_api.paper_autocomplete.fn(None, query="graph neural")

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/paper/autocomplete",
        params={"query": "graph neural"},
    )


async def test_snippet_search_happy_path(mock_make_request):
    payload = {"data": [{"snippet": {"text": "attention is all you need"}}]}
    mock_make_request.install(papers_api).queue_responses(payload)

    result = await papers_api.snippet_search.fn(
        None,
        query="transformer attention",
        fields=["snippet.text", "paper.title"],
        limit=5,
        paper_ids=["p1", "p2"],
        authors=["Author One", "Author Two"],
        min_citation_count=50,
        inserted_before="2025-01-01",
        publication_date_or_year="2020-01-01:2024-12-31",
        year="2020-2024",
        venue=["NeurIPS", "ICML"],
        fields_of_study=["Computer Science", "Mathematics"],
    )

    assert result == payload
    assert_single_call(
        mock_make_request,
        endpoint="/snippet/search",
        params={
            "query": "transformer attention",
            "limit": 5,
            "fields": "snippet.text,paper.title",
            "paperIds": "p1,p2",
            "authors": "Author One,Author Two",
            "minCitationCount": 50,
            "insertedBefore": "2025-01-01",
            "publicationDateOrYear": "2020-01-01:2024-12-31",
            "year": "2020-2024",
            "venue": "NeurIPS,ICML",
            "fieldsOfStudy": "Computer Science,Mathematics",
        },
    )


async def test_snippet_search_author_count_validation(mock_make_request):
    result = await papers_api.snippet_search.fn(
        None,
        query="transformer attention",
        authors=[f"Author {index}" for index in range(11)],
    )

    assert_validation_error(result, "Cannot filter by more than 10 authors", {"max_authors": 10})
    assert mock_make_request.calls == []


async def test_snippet_search_timeout_passthrough(mock_make_request, mock_error_response):
    error = mock_error_response(timeout=True)
    mock_make_request.install(papers_api).queue_responses(error)

    result = await papers_api.snippet_search.fn(
        None,
        query="transformer attention",
        fields=["snippet.text", "paper.title"],
        limit=5,
        paper_ids=["p1", "p2"],
        authors=["Author One", "Author Two"],
        min_citation_count=50,
        inserted_before="2025-01-01",
        publication_date_or_year="2020-01-01:2024-12-31",
        year="2020-2024",
        venue=["NeurIPS", "ICML"],
        fields_of_study=["Computer Science", "Mathematics"],
    )

    assert result == error
    assert_single_call(
        mock_make_request,
        endpoint="/snippet/search",
        params={
            "query": "transformer attention",
            "limit": 5,
            "fields": "snippet.text,paper.title",
            "paperIds": "p1,p2",
            "authors": "Author One,Author Two",
            "minCitationCount": 50,
            "insertedBefore": "2025-01-01",
            "publicationDateOrYear": "2020-01-01:2024-12-31",
            "year": "2020-2024",
            "venue": "NeurIPS,ICML",
            "fieldsOfStudy": "Computer Science,Mathematics",
        },
    )



import unittest
import asyncio
from typing import Optional, List, Dict
import random
import pytest

from .test_utils import make_request, create_error_response, ErrorType, Config

pytestmark = pytest.mark.live

class TestPaperTools(unittest.TestCase):
    def setUp(self):
        """Set up test environment"""
        # Create event loop for async tests
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        
        # Sample paper IDs for testing
        self.sample_paper_id = "649def34f8be52c8b66281af98ae884c09aef38b"
        self.sample_paper_ids = [
            self.sample_paper_id,
            "ARXIV:2106.15928"
        ]

    def tearDown(self):
        """Clean up after tests"""
        self.loop.close()

    def run_async(self, coro):
        """Helper to run async functions in tests"""
        return self.loop.run_until_complete(coro)

    async def async_test_with_delay(self, endpoint: str, **kwargs):
        """Helper to run async tests with delay to handle rate limiting"""
        await asyncio.sleep(random.uniform(5, 8))  # Random initial delay
        
        max_retries = 5
        base_delay = 8
        
        for attempt in range(max_retries):
            result = await make_request(endpoint, **kwargs)
            if not isinstance(result, dict) or "error" not in result:
                return result
                
            if result["error"]["type"] == "rate_limit":
                delay = base_delay * (2 ** attempt) + random.uniform(0, 2)  # Add jitter
                await asyncio.sleep(delay)
                continue
            else:
                return result
                
        return result  # Return last result if all retries failed

    @classmethod
    def setUpClass(cls):
        """Set up class-level test environment"""
        # Add initial delay before any tests run
        asyncio.get_event_loop().run_until_complete(asyncio.sleep(10))

    def test_paper_relevance_search(self):
        """Test paper relevance search functionality"""
        # Test basic search
        result = self.run_async(self.async_test_with_delay(
            "paper/search",  # Remove leading slash
            params={
                "query": "quantum computing",
                "fields": "title,abstract,year"
            }
        ))
        self.assertNotIn("error", result)
        self.assertIn("data", result)
        self.assertIn("total", result)
        
        # Test with filters
        result = self.run_async(self.async_test_with_delay(
            "paper/search",
            params={
                "query": "machine learning",
                "fields": "title,year",
                "minCitationCount": 100,
                "year": "2020-2023"
            }
        ))
        self.assertNotIn("error", result)
        self.assertIn("data", result)

    def test_paper_bulk_search(self):
        """Test paper bulk search functionality"""
        result = self.run_async(self.async_test_with_delay(
            "paper/search/bulk",  # Remove leading slash
            params={
                "query": "neural networks",
                "fields": "title,year,authors",
                "sort": "citationCount:desc"
            }
        ))
        self.assertNotIn("error", result)
        self.assertIn("data", result)

    def test_paper_details(self):
        """Test paper details functionality"""
        result = self.run_async(self.async_test_with_delay(
            f"paper/{self.sample_paper_id}",  # Remove leading slash
            params={
                "fields": "title,abstract,year,authors"
            }
        ))
        self.assertNotIn("error", result)
        self.assertIn("paperId", result)
        self.assertIn("title", result)

    def test_paper_batch_details(self):
        """Test batch paper details functionality"""
        result = self.run_async(self.async_test_with_delay(
            "paper/batch",  # Remove leading slash
            method="POST",
            params={"fields": "title,year,authors"},
            json={"ids": self.sample_paper_ids}
        ))
        self.assertNotIn("error", result)
        self.assertTrue(isinstance(result, list))
        self.assertEqual(len(result), len(self.sample_paper_ids))

if __name__ == '__main__':
    unittest.main()



"""
Author-related API endpoints for the Semantic Scholar API.
"""

from typing import Dict, List, Optional

from fastmcp import Context

from ..config import ErrorType
from ..core.client import S2Client, make_compat_client
from ..core.exceptions import S2ApiError, S2Error, S2ValidationError
from ..core.requests import (
    AuthorBatchDetailsRequest,
    AuthorDetailsRequest,
    AuthorPapersRequest,
    AuthorSearchRequest,
)
from ..mcp import mcp
from ..utils.errors import create_error_response, s2_exception_to_error_response
from ..utils.http import make_request


def _client() -> S2Client:
    return make_compat_client(make_request)


@mcp.tool()
async def author_search(
    context: Context,
    query: str,
    fields: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 100,
) -> Dict:
    try:
        request = AuthorSearchRequest(
            query=query,
            fields=fields,
            offset=offset,
            limit=limit,
        )
        return await _client().search_authors(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def author_details(
    context: Context,
    author_id: str,
    fields: Optional[List[str]] = None,
) -> Dict:
    try:
        request = AuthorDetailsRequest(author_id=author_id, fields=fields)
        return await _client().get_author(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Author not found",
                {"author_id": author_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def author_papers(
    context: Context,
    author_id: str,
    fields: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 100,
) -> Dict:
    try:
        request = AuthorPapersRequest(
            author_id=author_id,
            fields=fields,
            offset=offset,
            limit=limit,
        )
        return await _client().get_author_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Author not found",
                {"author_id": author_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def author_batch_details(
    context: Context,
    author_ids: List[str],
    fields: Optional[str] = None,
) -> Dict:
    try:
        request = AuthorBatchDetailsRequest(author_ids=author_ids, fields=fields)
        return await _client().batch_authors(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)



#!/usr/bin/env python3
"""
Audit local field allowlists against the live Semantic Scholar API spec.
"""

from __future__ import annotations

import sys
from pathlib import Path
from typing import Any

import httpx

ROOT = Path(__file__).resolve().parents[1]
if str(ROOT) not in sys.path:
    sys.path.insert(0, str(ROOT))

from semantic_scholar.config import (  # noqa: E402
    AuthorDetailFields,
    CitationReferenceFields,
    PaperFields,
)

SPEC_URLS = [
    "https://api.semanticscholar.org/graph/v1/swagger",
    "https://api.semanticscholar.org/graph/v1/openapi.json",
    "https://api.semanticscholar.org/graph/v1/swagger.json",
]
TIMEOUT_SECONDS = 20.0


class SpecFormatError(Exception):
    """Raised when the upstream spec shape is not recognized."""


def fetch_spec() -> dict[str, Any] | None:
    errors: list[str] = []
    try:
        with httpx.Client(timeout=TIMEOUT_SECONDS, follow_redirects=True) as client:
            for url in SPEC_URLS:
                try:
                    response = client.get(url, headers={"Accept": "application/json"})
                    response.raise_for_status()
                    payload = response.json()
                    if not isinstance(payload, dict):
                        errors.append(f"{url}: expected JSON object, got {type(payload).__name__}")
                        continue
                    return payload
                except (httpx.HTTPError, ValueError) as exc:
                    errors.append(f"{url}: {exc}")
    except Exception as exc:  # pragma: no cover - final safety net for transient client issues
        print(f"Warning: unable to initialize HTTP client: {exc}")
        return None

    print("Warning: unable to fetch Semantic Scholar API spec.")
    for error in errors:
        print(f"  - {error}")
    return None


def build_registry(spec: dict[str, Any]) -> dict[str, dict[str, Any]]:
    definitions = spec.get("definitions")
    if isinstance(definitions, dict):
        return {name: schema for name, schema in definitions.items() if isinstance(schema, dict)}

    components = spec.get("components")
    if isinstance(components, dict):
        schemas = components.get("schemas")
        if isinstance(schemas, dict):
            return {name: schema for name, schema in schemas.items() if isinstance(schema, dict)}

    raise SpecFormatError("spec does not contain Swagger definitions or OpenAPI components.schemas")


def ref_name(ref: str) -> str:
    return ref.rsplit("/", 1)[-1]


def collect_properties(
    schema: dict[str, Any] | None,
    registry: dict[str, dict[str, Any]],
    seen_refs: set[str] | None = None,
) -> dict[str, Any]:
    if not isinstance(schema, dict):
        return {}

    seen_refs = seen_refs or set()
    properties: dict[str, Any] = {}

    ref = schema.get("$ref")
    if isinstance(ref, str):
        if ref in seen_refs:
            return {}
        target = registry.get(ref_name(ref))
        if isinstance(target, dict):
            properties.update(collect_properties(target, registry, seen_refs | {ref}))

    all_of = schema.get("allOf")
    if isinstance(all_of, list):
        for part in all_of:
            properties.update(collect_properties(part, registry, seen_refs))

    schema_properties = schema.get("properties")
    if isinstance(schema_properties, dict):
        properties.update(schema_properties)

    return properties


def find_schemas(
    registry: dict[str, dict[str, Any]],
    candidates: list[str],
) -> list[dict[str, Any]]:
    lower_names = {name.lower(): name for name in registry}
    matches: list[dict[str, Any]] = []
    seen_names: set[str] = set()

    for candidate in candidates:
        matched_name = registry.get(candidate)
        if matched_name is None:
            actual_name = lower_names.get(candidate.lower())
            if actual_name is None:
                continue
            matched_name = registry[actual_name]
            schema_name = actual_name
        else:
            schema_name = candidate

        if schema_name in seen_names:
            continue
        seen_names.add(schema_name)
        matches.append(matched_name)

    return matches


def property_names_for_schemas(
    registry: dict[str, dict[str, Any]],
    candidates: list[str],
) -> set[str]:
    names: set[str] = set()
    schemas = find_schemas(registry, candidates)
    for schema in schemas:
        names.update(collect_properties(schema, registry).keys())
    return names


def extract_upstream_fields(spec: dict[str, Any]) -> dict[str, set[str]]:
    registry = build_registry(spec)

    paper_fields = property_names_for_schemas(
        registry,
        ["FullPaper", "PaperWithLinks", "BasePaper"],
    )
    if not paper_fields:
        raise SpecFormatError("unable to find paper schema definitions")

    author_schemas = find_schemas(
        registry,
        ["AuthorWithPapers", "AuthorDetail", "Author"],
    )
    author_fields: set[str] = set()
    paper_subfields: set[str] = set()
    for schema in author_schemas:
        properties = collect_properties(schema, registry)
        author_fields.update(properties.keys())
        papers_property = properties.get("papers")
        if isinstance(papers_property, dict):
            paper_subfields.update(
                collect_properties(papers_property.get("items"), registry).keys()
            )
    if not author_fields:
        raise SpecFormatError("unable to find author schema definitions")
    author_fields.update({f"papers.{name}" for name in paper_subfields})

    citation_fields: set[str] = set()
    citation_schemas = find_schemas(registry, ["Citation", "Reference", "CitationReference"])
    for schema in citation_schemas:
        properties = collect_properties(schema, registry)
        for name, property_schema in properties.items():
            if name in {"citingPaper", "citedPaper", "paper"}:
                citation_fields.update(collect_properties(property_schema, registry).keys())
            else:
                citation_fields.add(name)
    if not citation_fields:
        raise SpecFormatError("unable to find citation/reference schema definitions")

    return {
        "PaperFields.VALID_FIELDS": paper_fields,
        "AuthorDetailFields.VALID_FIELDS": author_fields,
        "CitationReferenceFields.VALID_FIELDS": citation_fields,
    }


def report_drift(name: str, local_fields: set[str], upstream_fields: set[str]) -> bool:
    missing = sorted(upstream_fields - local_fields)
    stale = sorted(local_fields - upstream_fields)
    if not missing and not stale:
        print(f"{name}: OK")
        return False

    print(f"{name}: DRIFT DETECTED")
    if missing:
        print("  New upstream fields missing locally:")
        for field in missing:
            print(f"    - {field}")
    if stale:
        print("  Local fields not present upstream:")
        for field in stale:
            print(f"    - {field}")
    return True


def main() -> int:
    spec = fetch_spec()
    if spec is None:
        return 0

    try:
        upstream_fields = extract_upstream_fields(spec)
    except SpecFormatError as exc:
        print(f"Warning: unexpected spec format: {exc}")
        return 0
    except Exception as exc:  # pragma: no cover - defensive fallback
        print(f"Warning: failed to parse Semantic Scholar API spec: {exc}")
        return 0

    local_field_sets = {
        "PaperFields.VALID_FIELDS": set(PaperFields.VALID_FIELDS),
        "AuthorDetailFields.VALID_FIELDS": set(AuthorDetailFields.VALID_FIELDS),
        "CitationReferenceFields.VALID_FIELDS": set(CitationReferenceFields.VALID_FIELDS),
    }

    drift_detected = False
    for name, local_fields in local_field_sets.items():
        drift_detected |= report_drift(name, local_fields, upstream_fields[name])

    if drift_detected:
        print("Semantic Scholar API spec drift detected.")
        return 1

    print("No Semantic Scholar API spec drift detected.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import unittest
import asyncio
from typing import Optional, List, Dict
import pytest

from .test_utils import make_request, create_error_response, ErrorType, Config

pytestmark = pytest.mark.live

class TestAuthorTools(unittest.TestCase):
    def setUp(self):
        """Set up test environment"""
        # Create event loop for async tests
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        
        # Sample author IDs for testing
        self.sample_author_id = "1741101"  # Andrew Ng
        self.sample_author_ids = [
            self.sample_author_id,
            "2061296"  # Yann LeCun
        ]

    def tearDown(self):
        """Clean up after tests"""
        self.loop.close()

    def run_async(self, coro):
        """Helper to run async functions in tests"""
        return self.loop.run_until_complete(coro)

    async def async_test_with_delay(self, coro):
        """Helper to run async tests with delay to handle rate limiting"""
        await asyncio.sleep(1)  # Add 1 second delay between tests
        return await coro

    def test_author_search(self):
        """Test author search functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            "/author/search",
            params={
                "query": "Andrew Ng",
                "fields": "name,affiliations,paperCount"
            }
        )))
        self.assertIn("data", result)
        self.assertIn("total", result)

    def test_author_details(self):
        """Test author details functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            f"/author/{self.sample_author_id}",
            params={
                "fields": "name,affiliations,paperCount,citationCount,hIndex"
            }
        )))
        self.assertIn("authorId", result)
        self.assertIn("name", result)

    def test_author_papers(self):
        """Test author papers functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            f"/author/{self.sample_author_id}/papers",
            params={
                "fields": "title,year,citationCount",
                "limit": 10
            }
        )))
        self.assertIn("data", result)
        self.assertIn("next", result)
        self.assertIn("offset", result)
        self.assertTrue(isinstance(result["data"], list))

    def test_author_batch_details(self):
        """Test batch author details functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            "/author/batch",
            method="POST",
            params={"fields": "name,affiliations,paperCount"},
            json={"ids": self.sample_author_ids}
        )))
        self.assertTrue(isinstance(result, list))
        self.assertEqual(len(result), len(self.sample_author_ids))

if __name__ == '__main__':
    unittest.main()



import unittest
import asyncio
import os
from typing import Optional, List, Dict
import pytest

from .test_utils import make_request, create_error_response, ErrorType, Config

pytestmark = pytest.mark.live

class TestRecommendationTools(unittest.TestCase):
    def setUp(self):
        """Set up test environment"""
        # API key is required for recommendations
        api_key = os.getenv("SEMANTIC_SCHOLAR_API_KEY")
        if not api_key or api_key.strip().lower() in ("", "none", "null", "false"):
            self.skipTest("SEMANTIC_SCHOLAR_API_KEY is required for recommendation tests")
        
        # Create event loop for async tests
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        
        # Sample paper IDs for testing (using full IDs)
        self.sample_paper_id = "204e3073870fae3d05bcbc2f6a8e263d9b72e776"  # "Attention is All You Need"
        self.positive_paper_ids = [
            self.sample_paper_id,
            "df2b0e26d0599ce3e70df8a9da02e51594e0e992"  # BERT
        ]
        self.negative_paper_ids = [
            "649def34f8be52c8b66281af98ae884c09aef38b"  # Different topic
        ]

    def tearDown(self):
        """Clean up after tests"""
        self.loop.close()

    def run_async(self, coro):
        """Helper to run async functions in tests"""
        return self.loop.run_until_complete(coro)

    async def async_test_with_delay(self, coro):
        """Helper to run async tests with delay to handle rate limiting"""
        await asyncio.sleep(1)  # Add 1 second delay between tests
        return await coro

    def test_paper_recommendations_single(self):
        """Test single paper recommendations functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            f"papers/forpaper/{self.sample_paper_id}",  # Using full paper ID
            params={
                "fields": "title,year"  # Minimal fields
            }
        )))
        self.assertIn("recommendedPapers", result)
        self.assertTrue(isinstance(result["recommendedPapers"], list))

    def test_paper_recommendations_multi(self):
        """Test multi-paper recommendations functionality"""
        result = self.run_async(self.async_test_with_delay(make_request(
            "papers",  # No leading slash
            method="POST",
            params={"fields": "title,year"},  # Minimal fields
            json={
                "positivePaperIds": self.positive_paper_ids,  # Changed key name to match API
                "negativePaperIds": self.negative_paper_ids
            }
        )))
        self.assertIn("recommendedPapers", result)
        self.assertTrue(isinstance(result["recommendedPapers"], list))

if __name__ == '__main__':
    unittest.main()



from collections import deque
from dataclasses import dataclass, field
from typing import Any

import pytest
import pytest_asyncio
from httpx import ASGITransport, AsyncClient

import semantic_scholar.bridge as bridge
from semantic_scholar.config import Config, ErrorType
from semantic_scholar.utils.errors import create_error_response


def pytest_configure(config):
    config.addinivalue_line("markers", "live: calls the real Semantic Scholar API")


@dataclass
class MockMakeRequestController:
    monkeypatch: pytest.MonkeyPatch
    calls: list[dict[str, Any]] = field(default_factory=list)
    queue: deque[Any] = field(default_factory=deque)
    default_response: Any = None

    def install(self, module):
        self.monkeypatch.setattr(module, "make_request", self.fake_make_request)
        return self

    def queue_responses(self, *items):
        self.queue.extend(items)

    async def fake_make_request(
        self,
        endpoint,
        params=None,
        api_key_override=None,
        method="GET",
        json=None,
        base_url=None,
    ):
        self.calls.append(
            {
                "endpoint": endpoint,
                "params": params,
                "api_key_override": api_key_override,
                "method": method,
                "json": json,
                "base_url": base_url,
            }
        )

        if self.queue:
            next_item = self.queue.popleft()
            if isinstance(next_item, Exception):
                raise next_item
            return next_item

        return self.default_response


@pytest.fixture
def mock_make_request(monkeypatch):
    return MockMakeRequestController(monkeypatch=monkeypatch)


@pytest.fixture
def mock_error_response():
    def factory(*, status_code=None, response="not found", retry_after="60", authenticated=True, timeout=False):
        if timeout:
            return create_error_response(
                ErrorType.TIMEOUT,
                f"Request timed out after {Config.TIMEOUT} seconds",
            )

        if status_code == 429:
            return create_error_response(
                ErrorType.RATE_LIMIT,
                "Rate limit exceeded. Consider using an API key for higher limits.",
                {
                    "status_code": 429,
                    "retry_after": retry_after,
                    "authenticated": authenticated,
                },
            )

        if status_code is not None:
            return create_error_response(
                ErrorType.API_ERROR,
                f"HTTP error: {status_code}",
                {
                    "status_code": status_code,
                    "response": response,
                },
            )

        return create_error_response(ErrorType.API_ERROR, "HTTP error: 500", {"status_code": 500, "response": response})

    return factory


@pytest_asyncio.fixture
async def bridge_client():
    async with AsyncClient(transport=ASGITransport(app=bridge.app), base_url="http://test") as client:
        yield client


@pytest.fixture
def auth_headers():
    return {"Authorization": "Bearer bridge-token"}



"""
Error handling utilities for the Semantic Scholar API Server.
"""

from typing import Any, Dict, Optional

from ..core.exceptions import S2ApiError, S2Error, S2RateLimitError, S2TimeoutError, S2ValidationError
from ..config import ErrorType


def create_error_response(
    error_type: ErrorType,
    message: str,
    details: Optional[Dict] = None
) -> Dict:
    """
    Create a standardized error response.

    Args:
        error_type: The type of error that occurred.
        message: A human-readable message describing the error.
        details: Optional additional details about the error.

    Returns:
        A dictionary with the error information.
    """
    return {
        "error": {
            "type": error_type.value,
            "message": message,
            "details": details or {}
        }
    }


def s2_exception_to_error_response(exc: S2Error) -> dict[str, Any]:
    if isinstance(exc, S2ValidationError):
        return create_error_response(ErrorType.VALIDATION, exc.message, exc.details)
    if isinstance(exc, S2RateLimitError):
        return create_error_response(
            ErrorType.RATE_LIMIT,
            exc.message,
            {
                "status_code": 429,
                "retry_after": exc.retry_after,
                "authenticated": exc.authenticated,
            },
        )
    if isinstance(exc, S2TimeoutError):
        return create_error_response(ErrorType.TIMEOUT, exc.message, {})
    if isinstance(exc, S2ApiError):
        details = {}
        if exc.status_code is not None:
            details["status_code"] = exc.status_code
        if exc.response_text is not None:
            details["response"] = exc.response_text
        return create_error_response(ErrorType.API_ERROR, exc.message, details)
    return create_error_response(ErrorType.API_ERROR, str(exc), {})



"""
API endpoints for the Semantic Scholar API Server.
"""

# Import all endpoints to make them available when importing the package
from .papers import (
    paper_relevance_search,
    paper_bulk_search,
    paper_title_search,
    paper_details,
    paper_batch_details,
    paper_authors,
    paper_citations,
    paper_references
)

from .authors import (
    author_search,
    author_details,
    author_papers,
    author_batch_details
)

from .recommendations import (
    get_paper_recommendations_single,
    get_paper_recommendations_multi
) 


"""
HTTP client compatibility helpers for the Semantic Scholar API Server.

Prefer importing `semantic_scholar.core` for new shared-client integrations.
"""

from typing import Any, Dict, Optional

from ..core.exceptions import S2Error
from ..core.transport import (
    RateLimiter,
    cleanup_client,
    default_transport,
    get_api_key,
    initialize_client,
    rate_limiter,
)
from .errors import s2_exception_to_error_response


async def make_request(
    endpoint: str,
    params: Optional[Dict[str, Any]] = None,
    api_key_override: Optional[str] = None,
    method: str = "GET",
    json: Any = None,
    base_url: Optional[str] = None,
) -> Any:
    """
    Compatibility wrapper around the shared core transport.

    Returns:
        The JSON response or an error response dictionary.
    """
    try:
        return await default_transport.request_json(
            endpoint,
            params=params,
            api_key_override=api_key_override,
            method=method,
            json=json,
            base_url=base_url,
        )
    except S2Error as exc:
        return s2_exception_to_error_response(exc)



"""
Paper-related API endpoints for the Semantic Scholar API.
"""

from typing import Dict, List, Optional

from fastmcp import Context

from ..config import ErrorType
from ..core.client import S2Client, make_compat_client
from ..core.exceptions import S2ApiError, S2Error, S2ValidationError
from ..core.requests import (
    PaperAutocompleteRequest,
    PaperAuthorsRequest,
    PaperBatchDetailsRequest,
    PaperBulkSearchRequest,
    PaperCitationsRequest,
    PaperDetailsRequest,
    PaperReferencesRequest,
    PaperRelevanceSearchRequest,
    PaperTitleSearchRequest,
    SnippetSearchRequest,
)
from ..mcp import mcp
from ..utils.errors import create_error_response, s2_exception_to_error_response
from ..utils.http import make_request


def _client() -> S2Client:
    return make_compat_client(make_request)


@mcp.tool()
async def paper_relevance_search(
    context: Context,
    query: str,
    fields: Optional[List[str]] = None,
    publication_types: Optional[List[str]] = None,
    open_access_pdf: bool = False,
    min_citation_count: Optional[int] = None,
    year: Optional[str] = None,
    venue: Optional[List[str]] = None,
    fields_of_study: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 10,
) -> Dict:
    try:
        request = PaperRelevanceSearchRequest(
            query=query,
            fields=fields,
            publication_types=publication_types,
            open_access_pdf=open_access_pdf,
            min_citation_count=min_citation_count,
            year=year,
            venue=venue,
            fields_of_study=fields_of_study,
            offset=offset,
            limit=limit,
        )
        return await _client().search_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_bulk_search(
    context: Context,
    query: Optional[str] = None,
    token: Optional[str] = None,
    fields: Optional[List[str]] = None,
    sort: Optional[str] = None,
    publication_types: Optional[List[str]] = None,
    open_access_pdf: bool = False,
    min_citation_count: Optional[int] = None,
    publication_date_or_year: Optional[str] = None,
    year: Optional[str] = None,
    venue: Optional[List[str]] = None,
    fields_of_study: Optional[List[str]] = None,
) -> Dict:
    try:
        request = PaperBulkSearchRequest(
            query=query,
            token=token,
            fields=fields,
            sort=sort,
            publication_types=publication_types,
            open_access_pdf=open_access_pdf,
            min_citation_count=min_citation_count,
            publication_date_or_year=publication_date_or_year,
            year=year,
            venue=venue,
            fields_of_study=fields_of_study,
        )
        return await _client().bulk_search_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_title_search(
    context: Context,
    query: str,
    fields: Optional[List[str]] = None,
    publication_types: Optional[List[str]] = None,
    open_access_pdf: bool = False,
    min_citation_count: Optional[int] = None,
    year: Optional[str] = None,
    venue: Optional[List[str]] = None,
    fields_of_study: Optional[List[str]] = None,
) -> Dict:
    try:
        request = PaperTitleSearchRequest(
            query=query,
            fields=fields,
            publication_types=publication_types,
            open_access_pdf=open_access_pdf,
            min_citation_count=min_citation_count,
            year=year,
            venue=venue,
            fields_of_study=fields_of_study,
        )
        return await _client().match_paper_title(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "No matching paper found",
                {"original_query": query},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_details(
    context: Context,
    paper_id: str,
    fields: Optional[List[str]] = None,
) -> Dict:
    try:
        request = PaperDetailsRequest(paper_id=paper_id, fields=fields)
        return await _client().get_paper(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Paper not found",
                {"paper_id": paper_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_batch_details(
    context: Context,
    paper_ids: List[str],
    fields: Optional[str] = None,
) -> Dict:
    try:
        request = PaperBatchDetailsRequest(paper_ids=paper_ids, fields=fields)
        return await _client().batch_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_authors(
    context: Context,
    paper_id: str,
    fields: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 100,
) -> Dict:
    try:
        request = PaperAuthorsRequest(
            paper_id=paper_id,
            fields=fields,
            offset=offset,
            limit=limit,
        )
        return await _client().get_paper_authors(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Paper not found",
                {"paper_id": paper_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_citations(
    context: Context,
    paper_id: str,
    fields: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 100,
) -> Dict:
    try:
        request = PaperCitationsRequest(
            paper_id=paper_id,
            fields=fields,
            offset=offset,
            limit=limit,
        )
        return await _client().get_paper_citations(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Paper not found",
                {"paper_id": paper_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_references(
    context: Context,
    paper_id: str,
    fields: Optional[List[str]] = None,
    offset: int = 0,
    limit: int = 100,
) -> Dict:
    try:
        request = PaperReferencesRequest(
            paper_id=paper_id,
            fields=fields,
            offset=offset,
            limit=limit,
        )
        return await _client().get_paper_references(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2ApiError as exc:
        if "404" in exc.message:
            return create_error_response(
                ErrorType.VALIDATION,
                "Paper not found",
                {"paper_id": paper_id},
            )
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def paper_autocomplete(
    context: Context,
    query: str,
) -> Dict:
    try:
        request = PaperAutocompleteRequest(query=query)
        return await _client().autocomplete_papers(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)


@mcp.tool()
async def snippet_search(
    context: Context,
    query: str,
    fields: Optional[List[str]] = None,
    limit: int = 10,
    paper_ids: Optional[List[str]] = None,
    authors: Optional[List[str]] = None,
    min_citation_count: Optional[int] = None,
    inserted_before: Optional[str] = None,
    publication_date_or_year: Optional[str] = None,
    year: Optional[str] = None,
    venue: Optional[List[str]] = None,
    fields_of_study: Optional[List[str]] = None,
) -> Dict:
    try:
        request = SnippetSearchRequest(
            query=query,
            fields=fields,
            limit=limit,
            paper_ids=paper_ids,
            authors=authors,
            min_citation_count=min_citation_count,
            inserted_before=inserted_before,
            publication_date_or_year=publication_date_or_year,
            year=year,
            venue=venue,
            fields_of_study=fields_of_study,
        )
        return await _client().search_snippets(request)
    except S2ValidationError as exc:
        return s2_exception_to_error_response(exc)
    except S2Error as exc:
        return s2_exception_to_error_response(exc)
